# Copyright 2021 Dallin Skouson, BYU CCL
# please see the BYU CCl SpyDrNet license file for terms of usage.


from spydrnet.parsers.verilog.tokenizer import VerilogTokenizer
import spydrnet.parsers.verilog.verilog_tokens as vt
from spydrnet.ir import Netlist, Library, Definition, Port, Cable, Instance, OuterPin
from spydrnet.plugins import namespace_manager
import spydrnet as sdn

from functools import reduce
import re


class VerilogParser:
    '''
    Parse verilog files into spydrnet.

    Higher level functions will always peek when deciding what lower level function to call.
    within your own function call next instead to keep the flow moving.

    the first token to expect in a function will be the token that starts that construct.
    '''

    #########################################################
    # Note to contributors
    #########################################################
    # I have tried to follow the convention that each function
    # parses all of the construct it is designed to parse
    # for example the parse module function will parse the
    # module keyword and then will call the parse
    # instance function. It will not consume any of the tokens
    # that belong to the instance instantations including the
    # semi colon. (it just uses peek)
    #
    # I would suggest following this convention even on constructs
    # where the first word is always the same.
    # the small overhead of using the peek function to not consume
    # the token has been well worth making it easier to maintain.
    # --Dallin

    #########################################################
    # helper classes
    #########################################################

    class BlackboxHolder:
        '''this is an internal class that helps manage
        modules that are instanced before they are declared'''

        def __init__(self):
            self.name_lookup = dict()
            self.defined = set()

        def get_blackbox(self, name):
            '''creates or returns the black box based on the name'''
            if name in self.name_lookup:
                return self.name_lookup[name]
            else:
                definition = sdn.Definition()
                definition.name = name
                self.name_lookup[name] = definition
                return definition

        def define(self, name):
            '''adds the name to the defined set'''
            self.defined.add(self.name_lookup[name])

        def get_undefined_blackboxes(self):
            '''return an iterable of all undefined blackboxes'''
            undef = set()
            for v in self.name_lookup.values():
                if v not in self.defined:
                    undef.add(v)
            return undef

    #######################################################
    # setup functions
    #######################################################
    @staticmethod
    def from_filename(filename):
        parser = VerilogParser()
        parser.filename = filename
        return parser

    @staticmethod
    def from_file_handle(file_handle):
        parser = VerilogParser()
        parser.filename = file_handle
        return parser

    def __init__(self):
        self.filename = None
        self.tokenizer = None

        self.netlist = None
        self.current_library = None
        self.current_definition = None
        self.current_instance = None

        self.primitives = None
        self.work = None
        self.assigns = None
        self.assignment_count = 0

        self.blackbox_holder = self.BlackboxHolder()

    def parse(self):
        ''' parse a verilog netlist represented by verilog file

            verilog_file can be a filename or stream
        '''
        self.initialize_tokenizer()
        ns_default = namespace_manager.default
        namespace_manager.default = "DEFAULT"
        self.parse_verilog()
        namespace_manager.default = ns_default
        self.tokenizer.__del__()
        return self.netlist

    def initialize_tokenizer(self):
        self.tokenizer = VerilogTokenizer(self.filename)

    def peek_token(self):
        token = self.peek_token_remove_comments()
        if token[0] == '`':
            t_split = token.split(maxsplit=1)
            if len(t_split) > 1 and t_split[0] in [vt.IFDEF]:
                while t_split[0] != vt.ENDIF:
                    token = self.next_token_remove_comments()
                    t_split = token.split(maxsplit=1)
                token = self.peek_token_remove_comments()
            if len(t_split) > 1 and t_split[0] == vt.DEFINE:
                assert False, self.error_string(
                    "define not supported", "assumes all macros are undefined", vt.DEFINE)
        return token

    def next_token(self):
        token = self.next_token_remove_comments()
        if token[0] == '`':
            t_split = token.split(maxsplit=1)
            if len(t_split) > 1 and t_split[0] in [vt.IFDEF]:
                while t_split[0] != vt.ENDIF:
                    token = self.next_token_remove_comments()
                    t_split = token.split(maxsplit=1)
                token = self.next_token_remove_comments()
            if len(t_split) > 1 and t_split[0] == vt.DEFINE:
                assert False, self.error_string(
                    "define not supported", "assumes all macros are undefined", vt.DEFINE)
        return token

    def peek_token_remove_comments(self):
        '''peeks from the tokenizer this wrapper function exists to skip comment tokens'''
        token = self.tokenizer.peek()
        while len(token) >= 2 and token[0] == "/" and (token[1] == "/" or token[1] == "*"):
            # this is a comment token skip it
            self.tokenizer.next()
            token = self.tokenizer.peek()
        return token

    def next_token_remove_comments(self):
        '''peeks from the tokenizer this wrapper function exists to skip comment tokens'''
        token = self.tokenizer.next()
        while len(token) >= 2 and (token[0:2] == vt.OPEN_LINE_COMMENT or token[0:2] == vt.OPEN_BLOCK_COMMENT):
            # this is a comment token, skip it
            token = self.tokenizer.next()
        return token

    #######################################################
    # parsing functions
    #######################################################

    def parse_verilog(self):
        self.netlist = sdn.Netlist()
        self.netlist.name = "SDN_VERILOG_NETLIST"
        self.work = self.netlist.create_library("work")
        self.primitives = self.netlist.create_library("SDN.verilog_primitives")
        self.current_library = self.work

        preprocessor_defines = set()
        star_properties = dict()
        time_scale = None
        primitive_cell = False

        while self.tokenizer.has_next():
            token = self.peek_token()
            if token.split(maxsplit=1)[0] == vt.CELL_DEFINE:
                primitive_cell = True
                self.current_library = self.primitives
                #token = token.split(maxsplit = 1)[1]
                token = self.next_token()
            elif token.split(maxsplit=1)[0] == vt.END_CELL_DEFINE:
                primitive_cell = False
                self.current_library = self.work
                #token = token.split(maxsplit = 1)[1]
                token = self.next_token()

            elif token == vt.MODULE:
                if primitive_cell:
                    self.parse_primitive()
                else:
                    self.parse_module()
                # go ahead and set the extra metadata that we collected to this point
                if time_scale is not None:
                    self.current_definition["VERILOG.TimeScale"] = time_scale
                if len(star_properties.keys()) > 0:
                    self.current_definition["VERILOG.InlineConstraints"] = star_properties
                    star_properties = dict()

            elif token == vt.PRIMITIVE:
                # self.parse_primitive()
                # if time_scale is not None:
                #     self.current_definition["VERILOG.TimeScale"] = time_scale
                # if len(star_properties.keys()) > 0:
                #     self.current_definition["VERILOG.InlineConstraints"] = star_properties
                #     star_properties = dict()
                star_properties = dict()
                while token != vt.END_PRIMITIVE:
                    token = self.next_token()

            elif token == vt.DEFINE:
                assert False, "Currently `define is not supported"
            elif token == vt.IFDEF:
                token = self.next_token()
                token = self.next_token()
                if token not in preprocessor_defines:
                    while token != vt.ENDIF:
                        token = self.next_token()
            elif token == vt.OPEN_PARENTHESIS:
                stars = self.parse_star_property()
                for k, v in stars.items():
                    star_properties[k] = v

            elif token.split(maxsplit=1)[0] == vt.TIMESCALE:
                token = self.next_token()
                time_scale = token.split(maxsplit=1)[1]
            else:
                pass
                assert False, self.error_string(
                    "something at the top level of the file", "got unexpected token", token)

        self.add_blackbox_definitions()

        return self.netlist

    def add_blackbox_definitions(self):
        self.current_library = self.primitives
        for d in self.blackbox_holder.get_undefined_blackboxes():
            d["VERILOG.primitive"] = True
            self.current_library.add_definition(d)

    def parse_primitive(self):
        '''similar to parse module but it will only look for the inputs and outputs to get an idea of how those things look'''

        token = self.next_token()
        assert token == vt.MODULE or token == vt.PRIMITIVE, self.error_string(
            vt.MODULE, "to begin module statement", token)
        token = self.next_token()
        assert vt.is_valid_identifier(token), self.error_string(
            "identifier", "not a valid module name", token)
        name = token

        definition = self.blackbox_holder.get_blackbox(name)
        self.blackbox_holder.define(name)
        self.current_library.add_definition(definition)
        self.current_definition = definition

        # uses the same header parser because the primitives and regular cells have the same header.
        self.parse_module_header()

        self.parse_primitive_body()

    def parse_primitive_body(self):
        ''' just look for port information, skip tasks and functions to help out.'''

        token = self.peek_token()

        while token != vt.END_MODULE and token != vt.END_PRIMITIVE:
            token = self.peek_token()
            if token == vt.FUNCTION:  # these constructs may contain input output or inout
                while token != vt.END_FUNCTION:
                    token = self.next_token()
            elif token == vt.TASK:  # these constructs may contain input output or inout
                while token != vt.END_TASK:
                    token = self.next_token()
            elif token in vt.PORT_DIRECTIONS:
                self.parse_port_declaration(dict())
            else:
                token = self.next_token()

    def parse_module(self):

        token = self.next_token()
        assert token == vt.MODULE, self.error_string(
            vt.MODULE, "to begin module statement", token)
        token = self.next_token()
        assert vt.is_valid_identifier(token), self.error_string(
            "identifier", "not a valid module name", token)
        name = token

        definition = self.blackbox_holder.get_blackbox(name)
        self.blackbox_holder.define(name)
        self.current_library.add_definition(definition)
        self.current_definition = definition
        self.assignment_count = 0
        if self.netlist.top_instance is None:
            self.netlist.top_instance = sdn.Instance()
            self.netlist.top_instance.name = definition.name + "_top"
            self.netlist.top_instance.reference = definition
            self.netlist.name = "SDN_VERILOG_NETLIST_" + definition.name

        self.parse_module_header()

        self.parse_module_body()

    def parse_module_header(self):
        '''parse a module header and add the parameter dictionary and port list to the current_definition'''
        token = self.peek_token()
        if token == "#":
            self.parse_module_header_parameters()

        token = self.peek_token()
        assert token == "(", self.error_string("(", "for port mapping", token)

        self.parse_module_header_ports()

        token = self.next_token()
        assert token == vt.CLOSE_PARENTHESIS, self.error_string(
            vt.CLOSE_PARENTHESIS, "to end the module ports in the header", token)

        token = self.next_token()
        assert token == vt.SEMI_COLON, self.error_string(
            vt.SEMI_COLON, "to end the module header section", token)

    def parse_module_header_parameters(self):
        '''parse a parameter block in a module header, add all parameters to the current definition'''
        token = self.next_token()
        assert token == vt.OCTOTHORP, self.error_string(
            vt.OCTOTHORP, "to begin parameter map", token)
        token = self.next_token()
        assert token == vt.OPEN_PARENTHESIS, self.error_string(
            vt.OPEN_PARENTHESIS, "to begin parameter map", token)

        token = self.next_token()

        parameter_dictionary = dict()

        while token != ")":
            # this is happening twice for all but the first one.. could simplify
            assert token == vt.PARAMETER, self.error_string(
                vt.PARAMETER, "parameter declaration", token)

            key = ""
            token = self.peek_token()
            if token == vt.OPEN_BRACKET:
                left, right = self.parse_brackets()
                if right != None:
                    key = "[" + str(left) + ":" + str(right) + "] "
                else:
                    key = "[" + str(left) + "] "

            token = self.next_token()
            assert vt.is_valid_identifier(token), self.error_string(
                'identifer', "in parameter list", token)
            key += token

            token = self.next_token()
            if key == vt.INTEGER:
                key += " " + token
                token = self.next_token()

            assert token == vt.EQUAL, self.error_string(
                vt.EQUAL, "in parameter list", token)

            token = self.next_token()
            # not really sure what to assert here.
            value = token

            parameter_dictionary[key] = value
            token = self.next_token()
            if token == vt.COMMA:  # just keep going
                token = self.next_token()
                assert token == vt.PARAMETER, self.error_string(
                    vt.PARAMETER, "after comma in parameter map", token)
            else:
                assert token == vt.CLOSE_PARENTHESIS, self.error_string(
                    vt.CLOSE_PARENTHESIS, "to end parameter declarations", token)

        self.set_definition_parameters(
            self.current_definition, parameter_dictionary)

    def parse_module_header_ports(self):
        '''parse port declarations in the module header and add them to the definition'''
        token = self.next_token()
        assert token == vt.OPEN_PARENTHESIS, self.error_string(
            vt.OPEN_PARENTHESIS, "to begin port declarations", token)

        token = self.peek_token()

        while token != ")":
            # the first token could be a name or input output or inout
            if token == ".":
                self.parse_module_header_port_alias()
            else:
                self.parse_module_header_port()
            token = self.peek_token()
            if token != vt.CLOSE_PARENTHESIS:
                assert token == vt.COMMA, self.error_string(
                    vt.COMMA, "to separate port declarations", token)
                token = self.next_token()  # consume the comma token
                token = self.peek_token()  # setup the next token

    def parse_module_header_port_alias(self):
        '''parse the port alias portion of the module header
        this parses the port alias section so that the port name is only a port and the mapped wires are the cables names that connect to that port.

        this requires that the cables names be kept in a dictionary to allow for setting the direction when the direction is given to the internal port names.

        example syntax
        .canale({\\canale[3] ,\\canale[2] ,\\canale[1] ,\\canale[0] }),'''

        token = self.next_token()
        assert token == vt.DOT, self.error_string(
            vt.DOT, "for port aliasing", token)
        token = self.next_token()
        assert vt.is_valid_identifier(token), self.error_string(
            "identifier", "for port in port aliasing", token)
        name = token

        token = self.next_token()
        assert token == vt.OPEN_PARENTHESIS, self.error_string(
            vt.OPEN_PARENTHESIS, "parethesis to enclose port aliasing", token)

        token = self.peek_token()
        if token == vt.OPEN_BRACE:
            wires = self.parse_cable_concatenation()
        else:
            cable, left, right = self.parse_variable_instantiation()
            wires = self.get_wires_from_cable(cable, left, right)

        token = self.next_token()
        assert token == vt.CLOSE_PARENTHESIS, self.error_string(
            vt.CLOSE_PARENTHESIS, "parethesis to end port aliasing construct", token)

        port = self.create_or_update_port(
            name, left_index=len(wires)-1, right_index=0)

        # connect the wires to the pins

        assert len(port.pins) == len(
            wires), "Internal Error: the pins in a created port and the number of wires the aliased cable do not match up"

        for i in range(len(port.pins)):
            wires[i].connect_pin(port.pins[i])

    def parse_cable_concatenation(self):
        '''parse a concatenation structure of cables, create the cables mentioned, and deal with indicies
        return a list of ordered wires that represents the cable concatenation
        example syntax
        {wire1, wire2, wire3, wire4}'''
        token = self.next_token()
        assert token == vt.OPEN_BRACE, self.error_string(
            vt.OPEN_BRACE, "to start cable concatenation", token)
        token = self.peek_token()
        wires = []
        while token != vt.CLOSE_BRACE:
            cable, left, right = self.parse_variable_instantiation()
            wires_temp = self.get_wires_from_cable(cable, left, right)
            for w in wires_temp:
                wires.append(w)
            token = self.next_token()
            if token != vt.COMMA:
                assert token == vt.CLOSE_BRACE, self.error_string(
                    vt.CLOSE_BRACE, "to end cable concatenation", token)

        return wires

    def parse_module_header_port(self):
        '''parse the port declaration in the module header'''
        token = self.peek_token()
        direction = None
        defining = False
        if token in vt.PORT_DIRECTIONS:
            token = self.next_token()
            direction = vt.string_to_port_direction(token)
            token = self.peek_token()
            defining = True

        left = None
        right = None
        if token == vt.OPEN_BRACKET:
            left, right = self.parse_brackets()

        token = self.next_token()
        assert vt.is_valid_identifier(token), self.error_string(
            "identifier", "for port declaration", token)
        name = token
        port = self.create_or_update_port(
            name, left_index=left, right_index=right, direction=direction, defining=defining)

        # get the left and right out of the port (in case we got more information out of an instance?)
        if left == None and right == None:
            left = port.lower_index + len(port.pins) - 1
            right = port.lower_index
            if not port.is_downto:
                temp = left
                left = right
                right = temp

        cable = self.create_or_update_cable(
            name, left_index=left, right_index=right, defining=defining)

        # wire together the cables and the port
        assert len(port.pins) == len(cable.wires), self.error_string(
            "the pins in a created port and the number of wires in it's cable do not match up", "wires: " + str(len(cable.wires)), "pins: " + str(len(port.pins)))
        for i in range(len(port.pins)):
            cable.wires[i].connect_pin(port.pins[i])

    def parse_module_body(self):
        '''
        parse through a module body

        module bodies consist of port declarations,
        wire and reg declarations
        and instantiations

        expects port declarations to start with the direction and then include the cable type if provided
        '''
        direction_tokens = [vt.INPUT, vt.OUTPUT, vt.INOUT]
        variable_tokens = [vt.WIRE, vt.REG, vt.TRI0, vt.TRI1]
        token = self.peek_token()
        params = dict()
        while token != vt.END_MODULE:
            if token in direction_tokens:
                self.parse_port_declaration(params)
                params = dict()
            elif token in variable_tokens:
                self.parse_cable_declaration(params)
                params = dict()
            elif token == vt.ASSIGN:
                o_cable, o_left, o_right, i_cable, i_left, i_right = self.parse_assign()
                self.connect_wires_for_assign(
                    o_cable, o_left, o_right, i_cable, i_left, i_right)
            elif token == vt.DEFPARAM:
                self.parse_defparam_parameters()
            elif vt.is_valid_identifier(token):
                self.parse_instantiation(params)
                params = dict()
            elif token == vt.OPEN_PARENTHESIS:
                stars = self.parse_star_property()
                for k, v in stars.items():
                    params[k] = v
            else:
                assert False, self.error_string(
                    "direction, reg, wire, star_properties, or instance identifier", "in module body", token)

            token = self.peek_token()

        token = self.next_token()
        assert token == vt.END_MODULE, self.error_string(
            vt.END_MODULE, "to end the module body", token)

    def parse_port_declaration(self, properties):
        '''parse the port declaration post port list.'''
        token = self.next_token()
        assert token in vt.PORT_DIRECTIONS, self.error_string(
            "direction keyword", "to define port", token)
        direction = vt.string_to_port_direction(token)

        token = self.peek_token()
        if token in [vt.REG, vt.WIRE]:
            var_type = token
            token = self.next_token()
        else:
            var_type = None

        token = self.peek_token()
        if token == vt.OPEN_BRACKET:
            left, right = self.parse_brackets()
        else:
            left = None
            right = None

        token = self.next_token()
        assert vt.is_valid_identifier(token), self.error_string(
            "port identifier", "identify port", token)
        names = []
        names.append(token)

        token = self.next_token()
        while token == vt.COMMA:
            token = self.next_token()
            names.append(token)
            token = self.next_token()

        assert token == vt.SEMI_COLON, self.error_string(
            vt.SEMI_COLON, "to end port declaration", token)

        for name in names:
            cable = self.create_or_update_cable(
                name, left_index=left, right_index=right, var_type=var_type, defining=True)

            port_list = self.get_all_ports_from_wires(
                self.get_wires_from_cable(cable, left, right))

            assert len(port_list) > 0, self.error_string(
                "port name defined in the module header", "to declare a port", cable.name)

            if len(port_list) > 1:
                for p in port_list:
                    port = self.create_or_update_port(
                        p.name, direction=direction)
                    port["VERILOG.InlineConstraints"] = properties
            else:
                port = self.create_or_update_port(port_list.pop(
                ).name, left_index=left, right_index=right, direction=direction, defining=True)

            if len(cable.wires) > 1:
                self.connect_resized_port_cable(cable, port)

    def parse_cable_declaration(self, properties):
        token = self.next_token()
        assert token in [vt.REG, vt.WIRE, vt.TRI0, vt.TRI1], self.error_string(
            "reg, tri1, tri0, or wire", "for cable declaration", token)
        var_type = token

        token = self.peek_token()
        if token == vt.OPEN_BRACKET:
            left, right = self.parse_brackets()
        else:
            left = None
            right = None

        token = self.next_token()
        assert vt.is_valid_identifier(token), self.error_string(
            "valid cable identifier", "identify the cable", token)
        name = token

        cable = self.create_or_update_cable(
            name, left_index=left, right_index=right, var_type=var_type)
        cable["VERILOG.InlineConstraints"] = properties

        token = self.next_token()
        assert token == vt.SEMI_COLON, self.error_string(
            vt.SEMI_COLON, "to end cable declaration", token)

    def parse_instantiation(self, properties):
        token = self.next_token()
        assert vt.is_valid_identifier(token), self.error_string(
            "module identifier", "for instantiation", token)
        def_name = token

        parameter_dict = dict()
        token = self.peek_token()
        if token == vt.OCTOTHORP:
            parameter_dict = self.parse_parameter_mapping()

        token = self.next_token()
        assert vt.is_valid_identifier(token), self.error_string(
            "instance name", "for instantiation", token)
        name = token

        token = self.peek_token()
        assert token == vt.OPEN_PARENTHESIS, self.error_string(
            vt.OPEN_PARENTHESIS, "to start port to cable mapping", token)

        instance = self.current_definition.create_child()
        self.current_instance = instance

        instance.name = name
        instance.reference = self.blackbox_holder.get_blackbox(def_name)
        instance["VERILOG.InlineConstraints"] = properties

        self.parse_port_mapping()

        self.set_instance_parameters(instance, parameter_dict)

        token = self.next_token()
        assert token == vt.SEMI_COLON, self.error_string(
            vt.SEMI_COLON, "to end instatiation", token)

    def parse_defparam_parameters(self):
        '''parse a defparam structure and add the parameters to the associated instance

        this looks like:

        defparam \\avs_s1_readdata[12]~output .bus_hold = "false"; //single backslash to escape name

        and must come after the associated instance (I'm not sure this is the verilog spec but
        it is the way quartus wrote my example and is much simpler)
        '''
        params = dict()
        token = self.next_token()
        assert token == vt.DEFPARAM, self.error_string(vt.DEFPARAM, "to being defparam statement", token)
        token = self.next_token()
        assert vt.is_valid_identifier(token), self.error_string("valid identifier", "of an instance to apply the defparam to", token)
        instance_name = token
        if self.current_instance.name == instance_name:
            instance = self.current_instance
        else:
            instance = next(self.current_definition.get_instances(instance_name), None)
            assert instance is not None, self.error_string("identifer of existing instance", "within the current definition", instance_name)
        token = self.next_token()
        assert token == vt.DOT, self.error_string(vt.DOT, "give separate parameter key from the instance name", token)
        token = self.next_token()
        key = token
        token = self.next_token()
        assert token == vt.EQUAL, self.error_string(vt.EQUAL, "separate the key from the value in a defparam statement", token)
        token = self.next_token()
        value = token
        token = self.next_token()
        assert token == vt.SEMI_COLON, self.error_string(vt.SEMI_COLON, "to end the defparam statement", token)
        self.set_instance_parameters(instance, params)


    def parse_parameter_mapping(self):
        params = dict()
        token = self.next_token()
        assert token == vt.OCTOTHORP, self.error_string(
            vt.OCTOTHORP, "to begin parameter mapping", token)

        token = self.next_token()
        assert token == vt.OPEN_PARENTHESIS, self.error_string(
            vt.OPEN_PARENTHESIS, "after # to begin parameter mapping", token)

        while token != vt.CLOSE_PARENTHESIS:
            k, v = self.parse_parameter_map_single()
            params[k] = v
            token = self.next_token()
            assert token in [vt.CLOSE_PARENTHESIS, vt.COMMA], self.error_string(
                vt.COMMA + " or " + vt.CLOSE_PARENTHESIS, "to separate parameters or end parameter mapping", token)

        assert token == vt.CLOSE_PARENTHESIS, self.error_string(
            vt.CLOSE_PARENTHESIS, "to terminate ", token)

        return params

    def parse_parameter_map_single(self):
        # syntax looks like .identifier(value)
        token = self.next_token()
        assert token == vt.DOT, self.error_string(
            vt.DOT, "to begin parameter mapping", token)

        token = self.next_token()
        assert vt.is_valid_identifier(token), self.error_string(
            "valid parameter identifier", "in parameter mapping", token)
        k = token

        token = self.next_token()
        assert token == vt.OPEN_PARENTHESIS, self.error_string(
            vt.OPEN_PARENTHESIS, "after identifier in parameter mapping", token)

        token = self.next_token()
        v = token

        token = self.next_token()
        assert token == vt.CLOSE_PARENTHESIS, self.error_string(
            vt.CLOSE_PARENTHESIS, "to close the parameter mapping value", token)

        return k, v

    def parse_port_mapping(self):
        token = self.next_token()
        assert token == vt.OPEN_PARENTHESIS, self.error_string(
            vt.OPEN_PARENTHESIS, "to start the port mapping", token)

        while token != vt.CLOSE_PARENTHESIS:
            self.parse_port_map_single()
            token = self.next_token()
            assert token in [vt.COMMA, vt.CLOSE_PARENTHESIS], self.error_string(
                vt.COMMA + " or " + vt.CLOSE_PARENTHESIS, "between port mapping elements or to end the port mapping", token)

    def parse_port_map_single(self):
        '''acutally does the mapping of the pins'''
        token = self.next_token()
        assert token == vt.DOT, self.error_string(
            vt.DOT, "to start a port mapping instance", token)

        token = self.next_token()
        assert vt.is_valid_identifier(token), self.error_string(
            "valid port identifier", "for port in instantiation port map", token)
        port_name = token

        token = self.next_token()
        assert token == vt.OPEN_PARENTHESIS, self.error_string(
            vt.OPEN_PARENTHESIS, "to encapsulate cable name in port mapping", token)

        token = self.peek_token()

        if token != vt.CLOSE_PARENTHESIS:

            if token == vt.OPEN_BRACE:
                wires = self.parse_cable_concatenation()
            else:
                cable, left, right = self.parse_variable_instantiation()
                wires = self.get_wires_from_cable(cable, left, right)

            pins = self.create_or_update_port_on_instance(
                port_name, len(wires))

            assert len(pins) >= len(wires), self.error_string(
                "pins length to match or exceed cable.wires length", "INTERNAL ERROR", str(len(pins)) + "!=" + str(len(wires)))

            # there can be unconnected pins at the end of the port.
            for i in range(len(wires)):
                wires[i].connect_pin(pins[i])

            token = self.next_token()

        else:
            # consume the )
            token = self.next_token()
            # the port is intentionally left unconnected.
            self.create_or_update_port_on_instance(port_name, 1)

        assert token == vt.CLOSE_PARENTHESIS, self.error_string(
            vt.CLOSE_PARENTHESIS, "to end cable name in port mapping", token)

    def parse_assign(self):
        token = self.next_token()
        assert token == vt.ASSIGN, self.error_string(
            vt.ASSIGN, "to begin assignment statement", token)
        l_cable, l_left, l_right = self.parse_variable_instantiation()
        token = self.next_token()
        assert token == vt.EQUAL, self.error_string(
            vt.EQUAL, "in assigment statment", token)
        r_cable, r_left, r_right = self.parse_variable_instantiation()
        token = self.next_token()
        assert token == vt.SEMI_COLON, self.error_string(
            vt.SEMI_COLON, "to terminate assign statement", token)

        return l_cable, l_left, l_right, r_cable, r_left, r_right

    def parse_variable_instantiation(self):
        '''parse the cable name and its indicies if any
        if we are in Intel land then 2 other things can happen.
        the "cable" is a constant,
            attach it to the \\<const0> or \\<const1> cable.
        the cable is inverted,
            create a new cable and an inverter block similar to the assign but with an inversion in the block
        '''
        token = self.next_token()

        if token[0] == "1":
            assert token[1] == vt.SINGLE_QUOTE, self.error_string(vt.SINGLE_QUOTE, "in the constant", token)
            assert token[2] == 'b', self.error_string('b', "in the constant", token)
            assert token[3] in ["0", "1", "x", "X", "z", "Z"], self.error_string("one of 0, 1, x, X, z, Z", "represent the constant value after '", token)
            name = "\\<const" + token[2] + "> "
        elif vt.is_numeric(token[0]):
            assert False, self.error_string("single bit constant", "multibit constants not supported", token)
        else:
            name = token
        assert vt.is_valid_identifier(name), self.error_string(
            "valid port identifier", "for port in instantiation port map", name)
        token = self.peek_token()
        left = None
        right = None
        if token == vt.OPEN_BRACKET:
            left, right = self.parse_brackets()

        cable = self.create_or_update_cable(
            name, left_index=left, right_index=right)

        return cable, left, right

    def parse_brackets(self):
        '''returns 2 integer values or 1 integer value and none'''
        token = self.next_token()
        assert token == vt.OPEN_BRACKET, self.error_string(
            "[", "to begin array slice", token)
        token = self.next_token()
        assert self.is_numeric(token), self.error_string(
            "number", "after [", token)
        left = int(token)
        token = self.next_token()
        if token == "]":
            return left, None
        else:
            assert(token == vt.COLON), self.error_string(
                "] or :", "in array slice", token)
            token = self.next_token()
            assert(self.is_numeric(token)), self.error_string(
                "number", "after : in array slice", token)
            right = int(token)
            token = self.next_token()
            assert token == vt.CLOSE_BRACKET, self.error_string(
                "]", "to terminate array slice", token)
            return left, right

    def parse_star_property(self):
        token = self.next_token()
        assert token == vt.OPEN_PARENTHESIS, self.error_string(
            vt.OPEN_PARENTHESIS, "to begin star property", token)
        token = self.next_token()
        assert token == vt.STAR, self.error_string(
            vt.STAR, "to begin star property", token)
        properties_dict = dict()
        token = self.next_token()
        while token != vt.STAR:
            assert vt.is_valid_identifier(token)
            key = token
            token = self.next_token()
            assert token in [vt.EQUAL, vt.STAR, vt.COMMA], self.error_string(
                vt.EQUAL + " or " + vt.STAR + " or " + vt.COMMA, "to set a star parameter", token)
            if token == vt.EQUAL:
                token = self.next_token()
                value = ""
                while token != vt.STAR and token != vt.COMMA:
                    value += token
                    token = self.next_token()
            else:
                value = None
            properties_dict[key] = value
            if token != vt.STAR:
                token = self.next_token()
        assert token == vt.STAR, self.error_string(
            vt.STAR, "to start the ending of a star property", token)
        token = self.next_token()
        assert token == vt.CLOSE_PARENTHESIS, self.error_string(
            vt.CLOSE_PARENTHESIS, "to end the star property", token)

        return properties_dict

    #######################################################
    # assignment helpers
    #######################################################

    def get_assignment_library(self):
        '''create or return a previously created assignment library'''
        if self.assigns == None:
            self.assigns = self.netlist.create_library(
                name="SDN_VERILOG_ASSIGNMENT")

        return self.assigns

    def get_assignment_definition(self, width):
        '''get the definition of the specified width for assignments'''
        proposed_name = "SDN_VERILOG_ASSIGNMENT_" + str(width)
        library = self.get_assignment_library()
        definition = next(library.get_definitions(proposed_name), None)
        if definition == None:
            definition = library.create_definition(name=proposed_name)
            in_port = definition.create_port("i")
            out_port = definition.create_port("o")

            in_port.create_pins(width)
            out_port.create_pins(width)
            in_port.direction = sdn.Port.Direction.IN
            out_port.direction = sdn.Port.Direction.OUT

            cable = definition.create_cable("through")
            cable.create_wires(width)
            for i in range(width):
                cable.wires[i].connect_pin(in_port.pins[i])
                cable.wires[i].connect_pin(out_port.pins[i])

        return definition

    def create_assignment_instance(self, width):
        '''create a new assign instance of the specified width on the current definition'''
        definition = self.get_assignment_definition(width)
        instance_name = definition.name + "_" + str(self.assignment_count)
        self.assignment_count += 1
        instance = self.current_definition.create_child(instance_name)
        instance.reference = definition
        return instance

    def connect_wires_for_assign(self, l_cable, l_left, l_right, r_cable, r_left, r_right):
        '''connect the wires in r_left to the wires in l_left'''

        out_wires = self.get_wires_from_cable(l_cable, l_left, l_right)
        in_wires = self.get_wires_from_cable(r_cable, r_left, r_right)

        # min because we don't need extra pins since only what can will assign.
        width = min(len(out_wires), len(in_wires))
        instance = self.create_assignment_instance(width)

        in_port = next(instance.reference.get_ports('i'), None)
        out_port = next(instance.reference.get_ports('o'), None)

        in_pins = self.get_pins_by_port_from_instance(instance, in_port)
        out_pins = self.get_pins_by_port_from_instance(instance, out_port)

        for i in range(width):
            out_wires[i].connect_pin(out_pins[i])
            in_wires[i].connect_pin(in_pins[i])

    #######################################################
    # helper functions
    #######################################################

    def get_pins_by_port_from_instance(self, instance, port):
        pin_lookup = instance.pins
        pins_out = []
        for p in port.pins:
            pins_out.append(pin_lookup[p])
        return pins_out

    def set_instance_parameters(self, instance, params):
        for k, v in params.items():
            #self.set_single_parameter(instance.reference, k, None)
            self.set_single_parameter(instance, k, v)

    def set_definition_parameters(self, definition, params):
        for k, v in params.items():
            self.set_single_parameter(definition, k, v)

    def set_single_parameter(self, var, k, v):
        if "VERILOG.Parameters" not in var:
            var["VERILOG.Parameters"] = dict()

        if k not in var["VERILOG.Parameters"] or var["VERILOG.Parameters"][k] is None:
            var["VERILOG.Parameters"][k] = v

    def get_all_ports_from_wires(self, wires):
        '''gets all ports associated with a set of wires'''
        ports = set()
        for w in wires:
            for p in w.pins:
                if isinstance(p, sdn.InnerPin):
                    ports.add(p.port)
        return ports

    def get_wires_from_cable(self, cable, left, right):
        wires = []

        if left != None and right != None:
            left = left - cable.lower_index
            right = right - cable.lower_index
            temp_wires = cable.wires[min(left, right): max(left, right) + 1]
            if left > right:
                temp_wires = reversed(temp_wires)

            for w in temp_wires:
                wires.append(w)

        elif left != None or right != None:
            if left != None:
                index = left - cable.lower_index
            else:
                index = right - cable.lower_index
            wires.append(cable.wires[index])

        else:
            for w in cable.wires:
                wires.append(w)

        return wires

    def convert_string_to_port_direction(self, token):
        if token == vt.INPUT:
            return sdn.Port.Direction.IN
        if token == vt.INOUT:
            return sdn.Port.Direction.INOUT
        if token == vt.OUTPUT:
            return sdn.Port.Direction.OUT
        else:
            return sdn.Port.Direction.UNDEFINED

    ########################################################################################
    # Port and cable creation and update managment
    ########################################################################################

    '''I'm handed a few different possible senarios

    module name(port1, port2,...);
    input [3:0] port1
    output [3:0] port2
    endmodule

    or

    module name
    (
        input [3:0] port1,
        output[3:0] port2,
        ...
    );

    additionally i need to be aware of the possibility that something like this happens

    module name(.port1({cable1, cable2}));
    input [1:0] cable1;
    output [1:0] cable2;
    '''

    def connect_resized_port_cable(self, resized_cable, resized_port):
        '''One to one connector. Don't use with alias statements. this expects that the given cable should completely fill the port...
        after a cable has been updated that is attached to a port it may need to update the port and reconnect the'''
        assert len(resized_cable.wires) == len(resized_port.pins), self.error_string("cable and port to have same size",
                                                                                     "to reconnect expanded cables and ports", "wires: " + str(len(resized_cable.wires)) + " pins: " + str(len(resized_port.pins)))
        for i in range(len(resized_port.pins)):
            # I think these should be lined up right?
            if resized_port.pins[i] not in resized_cable.wires[i].pins:
                resized_cable.wires[i].connect_pin(resized_port.pins[i])

    def create_or_update_cable(self, name, left_index=None, right_index=None, var_type=None, defining=False):
        cable_generator = self.current_definition.get_cables(name)
        cable = next(cable_generator, None)
        if cable == None:
            cable = self.current_definition.create_cable()
            self.populate_new_cable(
                cable, name, left_index, right_index, var_type)
            return cable

        assert cable.name == name

        cable_lower = cable.lower_index
        # -1 so that it is the same number if the width is 1
        cable_upper = cable.lower_index + len(cable.wires) - 1

        if left_index is not None and right_index is not None:
            in_lower = min(left_index, right_index)
            in_upper = max(left_index, right_index)
        elif left_index is not None:
            in_lower = left_index
            in_upper = left_index
        elif right_index is not None:
            in_upper = right_index
            in_lower = right_index
        else:
            in_upper = None
            in_lower = None

        if defining and in_lower is not None:  # if the cable width is being defined then recenter the cable
            cable.lower_index = in_lower
            cable_lower = cable.lower_index
            cable_upper = cable.lower_index + len(cable.wires) - 1

        if in_upper is not None and in_lower is not None:

            if in_lower < cable_lower:
                prepend = cable_lower - in_lower
                self.prepend_wires(cable, prepend)
            if in_upper > cable_upper:
                postpend = in_upper - cable_upper
                self.postpend_wires(cable, postpend)

        if var_type is not None:
            cable["VERILOG.CableType"] = var_type

        return cable

    def populate_new_cable(self, cable, name, left_index, right_index, var_type):
        cable.name = name
        if left_index is not None and right_index is not None:
            cable.is_downto = right_index <= left_index
            cable.create_wires(max(left_index, right_index) -
                               min(left_index, right_index) + 1)
            cable.lower_index = min(left_index, right_index)
        elif left_index is not None:
            cable.lower_index = left_index
            cable.create_wire()
        elif right_index is not None:
            cable.lower_index = right_index
            cable.create_wire()
        else:
            cable.lower_index = 0
            cable.create_wire()

        if var_type:
            cable["VERILOG.CableType"] = var_type

        return cable

    def prepend_wires(self, cable, count):
        orig_count = len(cable.wires)
        cable.create_wires(count)
        cable.wires = cable.wires[orig_count:] + cable.wires[:orig_count]
        cable.lower_index = cable.lower_index - count

    def postpend_wires(self, cable, count):
        cable.create_wires(count)

    def create_or_update_port_on_instance(self, name, width):
        '''returns the set of pins associated with the port on the instance'''
        pins = []
        port = self.create_or_update_port(
            name, left_index=width - 1, right_index=0, definition=self.current_instance.reference)
        for pin in self.current_instance.pins:
            if pin.inner_pin in port.pins:
                pins.append(pin)
        return pins

    def create_or_update_port(self, name, left_index=None, right_index=None, direction=None, definition=None, defining=False):
        if definition == None:
            definition = self.current_definition

        port_generator = definition.get_ports(name)
        port = next(port_generator, None)
        if port == None:
            port = definition.create_port()
            self.populate_new_port(
                port, name, left_index, right_index, direction)
            return port

        assert port.name == name

        # figure out what we need to do with the indicies

        port_lower = port.lower_index
        # -1 so that it is the same number if the width is 1
        port_upper = port.lower_index + len(port.pins) - 1

        if left_index is not None and right_index is not None:
            in_lower = min(left_index, right_index)
            in_upper = max(left_index, right_index)
        elif left_index is not None:
            in_lower = left_index
            in_upper = left_index
        elif right_index is not None:
            in_upper = right_index
            in_lower = right_index
        else:
            in_upper = None
            in_lower = None

        if defining and in_lower is not None:  # if the cable width is being defined then recenter the cable
            port.lower_index = in_lower
            port_lower = port.lower_index
            port_upper = port.lower_index + len(port.pins) - 1

        if in_upper is not None and in_lower is not None:

            if in_lower < port_lower:
                prepend = port_lower - in_lower
                self.prepend_pins(port, prepend)
            if in_upper > port_upper:
                postpend = in_upper - port_upper
                self.postpend_pins(port, postpend)

        if direction is not None:
            port.direction = direction

        return port

    def populate_new_port(self, port, name, left_index, right_index, direction):
        port.name = name
        if left_index is not None and right_index is not None:
            port.is_downto = right_index <= left_index
            port.create_pins(max(left_index, right_index) -
                             min(left_index, right_index) + 1)
            port.lower_index = min(left_index, right_index)
        elif left_index is not None:
            port.lower_index = left_index
            port.create_pin()
        elif right_index is not None:
            port.lower_index = right_index
            port.create_pin()
        else:
            port.lower_index = 0
            port.create_pin()

        if direction is not None:
            port.direction = direction

        return port

    def prepend_pins(self, port, count):
        orig_count = len(port.pins)
        port.create_pins(count)
        port.pins = port.pins[orig_count:] + port.pins[:orig_count]
        port.lower_index = port.lower_index - count

    def postpend_pins(self, port, count):
        port.create_pins(count)

    def is_numeric(self, token):
        first = True
        for c in token:
            if first:
                first = False
                if c == "-":
                    continue
            if c not in vt.NUMBERS:
                return False
        return True

    def is_alphanumeric(self, token):
        for c in token:
            if c not in vt.NUMBERS and c not in vt.LETTERS:
                return False
        return True

    def error_string(self, expected, why, result):
        '''put in the expectation and then the reason or location and the actual result'''
        return "expected " + str(expected) + " " + why + " but got " + str(result) + " Line: " + str(self.tokenizer.line_number)



import unittest
import spydrnet as sdn


class TestGetNetlists(unittest.TestCase):
    def test_parameter_checking(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        library.name = "MY_LIB"
        self.assertRaises(TypeError, sdn.get_netlists, netlist, "MY_LIB", patterns="MY_LIB")
        self.assertRaises(TypeError, sdn.get_netlists, netlist, "MY_LIB", unsupported_keyword=None)
        self.assertRaises(TypeError, sdn.get_netlists, None, "MY_LIB")
        self.assertRaises(TypeError, sdn.get_netlists, [None, netlist], "MY_LIB")

    def test_collection(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        library.name = "MY_LIB"
        netlists = list(sdn.get_netlists([netlist, netlist]))
        self.assertEqual(len(netlists), 1)

    def test_pin_and_wire_chain(self):
        netlist = sdn.Netlist()
        netlist.name = "MY_NETLIST"
        library = netlist.create_library()
        definition = library.create_definition()
        port = definition.create_port()
        pin = port.create_pin()
        cable = definition.create_cable()
        wire = cable.create_wire()
        instance = sdn.Instance()
        instance.name = "MY_INST"
        instance.reference = definition
        netlist.top_instance = instance

        netlist_lookup = next(wire.get_netlists(), None)
        self.assertIs(netlist, netlist_lookup)

        cable.remove_wire(wire)
        netlist_lookup = next(sdn.get_netlists(wire), None)
        self.assertIsNone(netlist_lookup)

        definition.remove_cable(cable)
        netlist_lookup = next(sdn.get_netlists(cable), None)
        self.assertIsNone(netlist_lookup)

        outer_pin = instance.pins[pin]
        netlist_lookup = next(sdn.get_netlists(outer_pin), None)
        self.assertIs(netlist, netlist_lookup)

        port.remove_pin(pin)
        netlist_lookup = next(sdn.get_netlists(pin), None)
        self.assertIsNone(netlist_lookup)

        self.assertIsNone(outer_pin.inner_pin)
        self.assertIsNone(outer_pin.instance)

        top_instance_href = next(sdn.get_hinstances(netlist.top_instance))
        netlist_lookup = next(sdn.get_netlists(top_instance_href))
        self.assertIs(netlist, netlist_lookup)

        definition.remove_port(port)
        netlist_lookup = next(sdn.get_netlists(port), None)
        self.assertIsNone(netlist_lookup)

        instance.reference = None
        netlist_lookup = next(sdn.get_netlists(instance), None)
        self.assertIsNone(netlist_lookup)

        library.remove_definition(definition)
        netlist_lookup = next(sdn.get_netlists(definition), None)
        self.assertIsNone(netlist_lookup)

        netlist.remove_library(library)
        netlist_lookup = next(sdn.get_netlists(library), None)
        self.assertIsNone(netlist_lookup)

        netlist_lookup = next(sdn.get_netlists(netlist, "MY_NETLIST"))
        self.assertTrue(netlist is netlist_lookup)



from spydrnet.plugins.namespace_manager.default_namespace import DefaultNamespace
from spydrnet.ir import Netlist, Library, Definition
import re
from spydrnet.plugins.namespace_manager.default_namespace import DefaultNamespace


class EdifNamespace(DefaultNamespace):

    @classmethod
    def is_name_of_element_valid(cls, element):
        """
        All names are valid in the default namespace. Always returns true.
        :param element:
        :return: True
        """
        if "EDIF.identifier" in element:
            identifier = element["EDIF.identifier"]
            return cls._check_EDIF_identifier(identifier)
        return True

    @classmethod
    def is_name_valid(cls, key, value):
        if key == "EDIF.identifier":
            return cls._check_EDIF_identifier(value)
        return True

    @classmethod
    def _check_EDIF_identifier(cls, identifier):
        identifier_length = len(identifier)
        starts_with_ampersand = identifier.startswith("&")
        if starts_with_ampersand:
            if identifier_length < 2 or identifier_length > 256:
                return False
            return bool(re.match(r"^&[0-9A-Za-z_]+$", identifier))
        else:
            if identifier_length < 1 or identifier_length > 255:
                return False
            if identifier[0].isalpha() is False:
                return False
            return bool(re.match(r"^[0-9A-Za-z_]+$", identifier))

    @classmethod
    def no_name_conflicts(cls, element):
        """
        Check to see if there are any name conflicts among the children
        :param element:
        :return:
        """
        namespace = set()
        edif_namespace = set()
        if isinstance(element, Netlist):
            for library in element.libraries:
                if ".NAME" in library:
                    name = library[".NAME"]
                    if name not in namespace:
                        namespace.add(name)
                    else:
                        return False
                if "EDIF.identifier" in library:
                    identifier = library["EDIF.identifier"].lower()
                    if identifier not in edif_namespace:
                        edif_namespace.add(identifier)
                    else:
                        return False
        elif isinstance(element, Library):
            for definition in element.definitions:
                if ".NAME" in definition:
                    name = definition[".NAME"]
                    if name not in namespace:
                        namespace.add(name)
                    else:
                        return False
                if "EDIF.identifier" in definition:
                    identifier = definition["EDIF.identifier"].lower()
                    if identifier not in edif_namespace:
                        edif_namespace.add(identifier)
                    else:
                        return False
        elif isinstance(element, Definition):
            for port in element.ports:
                if ".NAME" in port:
                    name = port[".NAME"]
                    if name not in namespace:
                        namespace.add(name)
                    else:
                        return False
                if "EDIF.identifier" in port:
                    identifier = port["EDIF.identifier"].lower()
                    if identifier not in edif_namespace:
                        edif_namespace.add(identifier)
                    else:
                        return False
            namespace.clear()
            for cable in element.cables:
                if ".NAME" in cable:
                    name = cable[".NAME"]
                    if name not in namespace:
                        namespace.add(name)
                    else:
                        return False
                if "EDIF.identifier" in cable:
                    identifier = cable["EDIF.identifier"].lower()
                    if identifier not in edif_namespace:
                        edif_namespace.add(identifier)
                    else:
                        return False
            namespace.clear()
            for instance in element.children:
                if ".NAME" in instance:
                    name = instance[".NAME"]
                    if name not in namespace:
                        namespace.add(name)
                    else:
                        return False
                if "EDIF.identifier" in instance:
                    identifier = instance["EDIF.identifier"].lower()
                    if identifier not in edif_namespace:
                        edif_namespace.add(identifier)
                    else:
                        return False
        return True

    def __init__(self):
        self.namespaces = dict()
        self.edif_namespaces = dict()

    def no_conflict(self, element, key, value):
        element_type = type(element)
        if key == ".NAME":
            if element_type in self.namespaces:
                namespace = self.namespaces[element_type]
                if value in namespace:
                    if namespace[value] != element:
                        return False
        elif key == "EDIF.identifier":
            if element_type in self.edif_namespaces:
                namespace = self.edif_namespaces[element_type]
                value_lower = value.lower()
                if value_lower in namespace:
                    if namespace[value_lower] != element:
                        return False
        return True

    def update(self, element, key, value):
        element_type = type(element)
        if key == ".NAME":
            if element_type not in self.namespaces:
                self.namespaces[element_type] = dict()
            namespace = self.namespaces[element_type]
            if ".NAME" in element:
                old_name = element[".NAME"]
                if old_name in namespace:
                    del namespace[old_name]
            namespace[value] = element
        elif key == "EDIF.identifier":
            if element_type not in self.edif_namespaces:
                self.edif_namespaces[element_type] = dict()
            namespace = self.edif_namespaces[element_type]
            if "EDIF.identifier" in element:
                old_name = element["EDIF.identifier"].lower()
                if old_name in namespace:
                    del namespace[old_name]
            namespace[value.lower()] = element

    def remove(self, element, key):
        element_type = type(element)
        if key == ".NAME":
            if element_type in self.namespaces:
                namespace = self.namespaces[element_type]
                if ".NAME" in element:
                    old_name = element[".NAME"]
                    if old_name in namespace:
                        del namespace[old_name]
        elif key == "EDIF.identifier":
            if element_type in self.edif_namespaces:
                namespace = self.edif_namespaces[element_type]
                if "EDIF.identifier" in element:
                    old_name = element["EDIF.identifier"]
                    if old_name in namespace:
                        del namespace[old_name.lower()]

    def lookup(self, element_type, key, value):
        if key == ".NAME":
            if element_type in self.namespaces:
                namespace = self.namespaces[element_type]
                return namespace.get(value, None)
        elif key == "EDIF.identifier":
            if element_type in self.edif_namespaces:
                namespace = self.edif_namespaces[element_type]
                return namespace.get(value.lower(), None)



import unittest

from spydrnet.clone import clone
from spydrnet.ir import *
from spydrnet.compare.compare_netlists import Comparer


class TestClone(unittest.TestCase):
    
    def _get_two_netlists(self):
        #create a test netlist
        nl = self._create_netlist()
        #create my own backup of the test netlist (just create another that is identical)
        nl2 = self._create_netlist()
        
        #both netlists are created and checkout as the same.
        return nl, nl2

    def _compare_netlists(self, n1, n2):
        comparer = Comparer(n1, n2)
        comparer.compare()

    def _create_netlist(self):
        netlist = Netlist()
        lib = netlist.create_library()
        defin = lib.create_definition()
        netlist.top_instance = Instance()
        netlist.top_instance.reference = defin
        port = defin.create_port()
        cable = defin.create_cable()
        instance = defin.create_child()
        defin2 = lib.create_definition()
        instance.reference = defin2
        wire = cable.create_wire()
        port.create_pins(2)
        port2 = defin2.create_port()
        port2.create_pins(1)
        for pin in port.pins:
            wire.connect_pin(pin)
        for pin in instance.pins:
            wire.connect_pin(pin)
        lib1 = netlist.create_library()
        lib2 = netlist.create_library()
        def1 = lib1.create_definition()
        def2 = lib2.create_definition()
        def3 = lib1.create_definition()
        def4 = lib2.create_definition()
        ins1 = def1.create_child()
        ins2 = def2.create_child()
        ins3 = def3.create_child()
        def4.create_child()
        ins1.reference = def2
        ins2.reference = def3
        ins3.reference = def4
        return netlist


    def create_and_clone_cable(self, wirecount, array, downto, index, key, value):
        definition = Definition()
        p1 = definition.create_cable()
        p1[key] = value
        p1.create_wires(wirecount)
        p1.is_array = array
        p1.is_downto = downto
        p1.lower_index = index
        p2 = clone(p1)
        assert p2.lower_index == index
        assert p2.is_downto == downto
        assert p2.is_array == array
        assert len(p2.wires) == wirecount
        assert p2[key] == value
        assert p2.definition == None
        for pin in p2.wires:
            assert pin.cable is p2
            assert len(pin.pins) == 0

    def test_cable_array(self):
        pincount = 10
        array = True
        downto = False
        index = 5
        key = "EDIF.identifier"
        value = "myPort"
        self.create_and_clone_cable(pincount, array, downto, index, key, value)

    def test_cable_not_array(self):
        pincount = 1
        array = False
        downto = True
        index = 0
        key = "Name"
        value = "myPortName"
        self.create_and_clone_cable(pincount, array, downto, index, key, value)        

    def test_cable_single_bit_array(self):
        pincount = 1
        array = True
        downto = False
        index = 0
        key = "garbage_key"
        value = "garbage_value"
        self.create_and_clone_cable(pincount, array, downto, index, key, value)        

    def test_definition(self):
        lib = Library()
        def1 = lib.create_definition()
        def2 = lib.create_definition()
        ins1 = def2.create_child()
        ins2 = def2.create_child()
        ins1.reference = def1
        ins2.reference = def1
        def2.create_port()
        def2.create_cable()
        def3 = clone(def2)
        assert def3.library == None
        assert len(def3.children) == len(def2.children)
        assert len(def3.cables) == len(def2.cables)
        assert len(def3.ports) == len(def2.ports)
        for p in def3.ports:
            assert p not in def2.ports
        for c in def3.cables:
            assert c not in def2.cables
        for c in def3.children:
            assert c not in def2.children
            assert c in def1.references
        assert len(def1.references) == 4

    def test_innerpin(self):
        port = Port()
        pin = port.create_pin()
        pin2 = pin.clone()
        assert pin2.wire == None
        assert pin2.port == None
        
    def test_instance(self):
        def1 = Definition()
        def2 = Definition()
        por2 = def2.create_port()
        por2.create_pins(5)
        child = def1.create_child()
        child.reference = def2
        inst2 = clone(child)
        assert inst2.parent == None
        assert inst2.reference is def2
        assert set(inst2.pins.keys()).difference(set(child.pins.keys())) == set()
        for v in inst2.pins.values():
            assert v not in child.pins.values()


    def test_library_instance_references(self):
        lib1 = Library()
        lib2 = Library()
        def1 = lib1.create_definition()
        def2 = lib1.create_definition()
        def3 = lib2.create_definition()
        ins2 = def1.create_child()
        ins3 = def1.create_child()
        ins2.reference = def2
        ins3.reference = def3
        lib3 = clone(lib1)
        assert lib3.netlist == None
        assert len(lib3.definitions) == len(lib1.definitions)
        for d in lib3.definitions:
            assert d not in lib1.definitions
        def1c = lib3.definitions[0]
        def2c = lib3.definitions[1]
        ins2c = def1c.children[0]
        ins3c = def1c.children[1]
        assert ins2c in def2c.references
        assert ins2c not in def2.references
        assert ins3c in def3.references
        assert ins2c.reference is def2c
        assert ins3c.reference is def3
        assert ins2 in def2.references
        assert ins2 not in def2c.references
        assert len(def3.references) == 2


    def test_library_definition_references(self):
        lib1 = Library()
        lib2 = Library()
        def1 = lib1.create_definition()
        def2 = lib2.create_definition()
        ins1 = def2.create_child()
        ins1.reference = def1
        lib3 = clone(lib1)
        def1c = lib3.definitions[0]
        assert len(def1c.references) == 0
        assert ins1.reference is def1

    def check(self, r1, e):
        if e is None:
            return
        assert e not in r1

    def not_among_all_references(self,r1, nl1):
        self.check(r1, nl1)
        self.check(r1, nl1.top_instance)
        for l in nl1.libraries:
            self.check(r1, l)
            for d in l.definitions:
                self.check(r1, d)
                for c in d.children:
                    self.check(r1, c)
                    self.check(r1, c.reference)
                    for k, v in c.pins.items():
                        self.check(r1, k)
                        self.check(r1, v)
                for r in d.references:
                    self.check(r1, r)
                for c in d.cables:
                    self.check(r1, c)
                    for w in c.wires:
                        self.check(r1, w)
                for p in d.ports:
                    self.check(r1, p)
                    for pi in p.pins:
                        self.check(r1, pi)

    def add_all_references(self,r1, nl1):
        r1.add(nl1)
        r1.add(nl1.top_instance)
        for l in nl1.libraries:
            r1.add(l)
            for d in l.definitions:
                r1.add(d)
                for c in d.children:
                    r1.add(c)
                    r1.add(c.reference)
                    for k, v in c.pins.items():
                        r1.add(k)
                        r1.add(v)
                for r in d.references:
                    r1.add(r)
                for c in d.cables:
                    r1.add(c)
                    for w in c.wires:
                        r1.add(w)
                for p in d.ports:
                    r1.add(p)
                    for pi in p.pins:
                        r1.add(pi)

    def check_overlap_references(self, nl1, nl2):
        r1 = set()
        self.add_all_references(r1, nl1)
        self.not_among_all_references(r1, nl2)
        
    def test_netlist_several_lib(self):
        netlist = Netlist()
        lib1 = netlist.create_library()
        lib2 = netlist.create_library()
        lib3 = netlist.create_library()
        lib4 = netlist.create_library()
        def1a = lib1.create_definition()
        def2a = lib2.create_definition()
        def3a = lib3.create_definition()
        def4a = lib4.create_definition()
        def1b = lib1.create_definition()
        def2b = lib2.create_definition()
        def3b = lib3.create_definition()
        def4b = lib4.create_definition()
        ins1a = def1a.create_child()
        ins2a = def2a.create_child()
        ins3a = def3a.create_child()
        ins4a = def4a.create_child()
        ins1b = def1b.create_child()
        ins2b = def2b.create_child()
        ins3b = def3b.create_child()
        def4b.create_child()
        ins1a.reference = def2a
        ins2a.reference = def3a
        ins3a.reference = def4a
        ins4a.reference = def1b
        ins1b.reference = def2b
        ins2b.reference = def3b
        ins3b.reference = def4b
        netlist2 = clone(netlist)
        self._compare_netlists(netlist, netlist2)
        self.check_overlap_references(netlist, netlist2)
        
    def test_netlist_change_top_instance(self):
        nl1 = Netlist()
        lib1 = nl1.create_library()
        def1 = lib1.create_definition()
        nl1.top_instance = Instance()
        nl1.top_instance.reference = def1
        nl1.top_instance = None
        nl2 = clone(nl1)
        self._compare_netlists(nl1, nl2)
        self.check_overlap_references(nl1, nl2)


    def test_library_change_top_instace(self):
        nl1 = Netlist()
        nl2 = Netlist()
        lib1 = nl1.create_library()
        def1 = lib1.create_definition()
        nl1.top_instance = Instance()
        nl1.top_instance.reference = def1
        nl1.top_instance = None
        lib2 = clone(lib1)
        nl2.add_library(lib2)
        self._compare_netlists(nl1, nl2)
        self.check_overlap_references(nl1, nl2)
        
    
    def test_netlist(self):
        nl1, nl2 = self._get_two_netlists()
        nl3 = clone(nl1)
        self._compare_netlists(nl1, nl3)
        self._compare_netlists(nl1, nl2)
        #now check that no references overlap.
        self.check_overlap_references(nl1, nl3)

    def test_netlist_empty_top_instance(self):
        nl1 = self._create_netlist()
        nl1.top_instance = None
        nl2 = clone(nl1)
        self._compare_netlists(nl1, nl2)
        self.check_overlap_references(nl1, nl2)

    def test_netlist_top_instance_instanced_elsewhere(self):
        nl1 = self._create_netlist()
        in1 = nl1.libraries[0].definitions[0].create_child()
        in1.reference = nl1.libraries[0].create_definition()
        nl1.top_instance = in1

        nl2 = clone(nl1)
        self._compare_netlists(nl1, nl2)
        self.check_overlap_references(nl1, nl2)

    def test_outerpin(self):
        op = OuterPin()
        wire = Wire()
        inner = InnerPin()
        op._wire = wire
        op._inner_pin = inner
        op2 = clone(op)
        assert op2.wire == None
        assert op2.instance == None
        assert op2.inner_pin == None

    def create_and_clone_port(self, pincount, direction, array, downto, index, key, value):
        definition = Definition()
        p1 = definition.create_port()
        p1[key] = value
        p1.create_pins(pincount)
        p1.direction = direction
        p1.is_array = array
        p1.is_downto = downto
        p1.lower_index = index
        p2 = clone(p1)
        assert p2.lower_index == index
        assert p2.is_downto == downto
        assert p2.is_array == array
        assert p2.direction == direction
        assert len(p2.pins) == pincount
        assert p2[key] == value
        assert p2.definition == None
        for pin in p2.pins:
            assert pin.port is p2
            assert pin.wire == None

    def test_port_array(self):
        pincount = 10
        direction = Port.Direction.IN
        array = True
        downto = False
        index = 5
        key = "EDIF.identifier"
        value = "myPort"
        self.create_and_clone_port(pincount, direction, array, downto, index, key, value)        
    
    def test_port_not_array(self):
        pincount = 1
        direction = Port.Direction.OUT
        array = False
        downto = True
        index = 0
        key = "Name"
        value = "myPortName"
        self.create_and_clone_port(pincount, direction, array, downto, index, key, value)        

    def test_port_single_bit_array(self):
        pincount = 1
        direction = Port.Direction.INOUT
        array = True
        downto = False
        index = 0
        key = "garbage_key"
        value = "garbage_value"
        self.create_and_clone_port(pincount, direction, array, downto, index, key, value)        

    def test_port_with_connectivity(self):
        pass #TODO

    def test_wire(self):
        cable = Cable()
        wire = cable.create_wire()
        wire2 = clone(wire)
        assert len(wire2.pins) == 0
        assert wire2.cable == None


from spydrnet.ir import InnerPin, OuterPin, Wire, Netlist, Library, Definition, Element, Instance,\
    Port, Cable
from spydrnet.util.hierarchical_reference import HRef
from spydrnet.util.selection import Selection
from spydrnet.global_state.global_service import lookup
from spydrnet.util.patterns import _is_pattern_absolute, _value_matches_pattern


def get_definitions(obj, *args, **kwargs):
    """
    get_definitions(obj, ...)

    Get definitions *within* an object.

    Parameters
    ----------
    obj : object, Iterable - required
        The object or objects associated with this query. Queries return a collection objects associated with the
        provided object or objects that match the query criteria. For example, `sdn.get_definitions(library, ...)` would
        return all of the definitions associated with the provided library that match the additional criteria.
    patterns : str, Iterable - optional, positional or named, default: wildcard
        The search patterns. Patterns can be a single string or an Iterable collection of strings. Patterns can be
        absolute or they can contain wildcards or regular expressions. If `patterns` is not provided, then it defaults
        to a wildcard. Patterns are queried against the object property value stored under a specified key. Fast lookups
        are only attempted on absolute patterns that are not regular expressions and contain no wildcards.
    key : str, optional, default: ".NAME"
        This is the key that controls which value is being searched.
    is_case : bool - optional, named, default: True
        Specify if patterns should be treated as case sensitive. Only applies to patterns. Does not alter fast lookup
        behavior (if namespace policy uses case insensitive indexing, this parameter will not prevent a fast lookup
        from returning a matching object even if the case is not an exact match).
    is_re: bool - optional, named, default: False
        Specify if patterns are regular expressions. If `False`, a pattern can still contain `*` and `?` wildcards. A
        `*` matches zero or more characters. A `?` matches upto a single character.
    recursive : bool - optional, default: False
        Specify if search should be recursive or not meaning that sub hierarchical instances within an instance are
        included or not.
    selection : Selection.{INSIDE, OUTSIDE}, default: INSIDE
        This parameter determines the instances that are returned based on the definition that is being searched. This
        parameter only applies to objects that are definitions. If the selection is "INSIDE" (default), then the
        function will return all of the instances that are inside the definition (i.e., the definition's children) that
        match the remainder of the search criteria. If the selection is "OUTSIDE", then the function will return all of
        the instances of the provided definition that match the remainder of the search criteria.
    filter : function
        This is a single input function that can be used to filter out unwanted virtual instances. If not specifed, all
        matching virtual instances are returned. Otherwise, virtual instances that cause the filter function to evaluate
        to true are the only items returned.
    
    Returns
    -------
    definitions : generator
        The definitions associated with a particular object or collection of objects.
    
    """
    # Check argument list
    if len(args) == 1 and 'patterns' in kwargs:
        raise TypeError("get_definitions() got multiple values for argument 'patterns'")
    if len(args) > 1 or any(x not in {'patterns', 'key', 'filter', 'is_case', 'is_re', 'selection', 'recursive'}
                            for x in kwargs):
        raise TypeError("Unknown usage. Please see help for more information.")

    # Default values
    selection = kwargs.get('selection', Selection.INSIDE)
    if isinstance(selection, str):
        if selection in Selection.__members__:
            selection = Selection[selection]
    if selection not in {Selection.INSIDE, Selection.OUTSIDE}:
        raise TypeError("selection must be '{}'".format("', '".join([Selection.INSIDE.name, Selection.OUTSIDE.name])))

    filter_func = kwargs.get('filter', lambda x: True)
    is_case = kwargs.get('is_case', True)
    is_re = kwargs.get('is_re', False)
    patterns = args[0] if len(args) == 1 else kwargs.get('patterns', ".*" if is_re else "*")
    key = kwargs.get('key', ".NAME")
    recursive = kwargs.get('recursive', False)

    if isinstance(obj, (Element, HRef)) is False:
        try:
            object_collection = list(iter(obj))
        except TypeError:
            object_collection = [obj]
    else:
        object_collection = [obj]
    if all(isinstance(x, (Element, HRef)) for x in object_collection) is False:
        raise TypeError("get_definitions() only supports netlists and libraries or a collection of them as the object "
                        "searched")

    if isinstance(patterns, str):
        patterns = (patterns,)

    return _get_definitions(object_collection, patterns, key, is_case, is_re, selection, recursive, filter_func)


def _get_definitions(object_collection, patterns, key, is_case, is_re, selection, recursive, filter_func):
    for result in filter(filter_func, _get_definitions_raw(object_collection, patterns, key, is_case, is_re,
                                                           selection, recursive)):
        yield result


def _get_definitions_raw(object_collection, patterns, key, is_case, is_re, selection, recursive):
    found = set()
    other_definitions = set()
    while object_collection:
        obj = object_collection.pop()
        if isinstance(obj, Library):
            if selection == Selection.INSIDE:
                for pattern in patterns:
                    pattern_is_absolute = _is_pattern_absolute(pattern, is_case, is_re)
                    if pattern_is_absolute:
                        result = lookup(obj, Definition, key, pattern)
                        if result is not None and result not in found:
                            found.add(result)
                            yield result
                    else:
                        for definition in obj.definitions:
                            value = definition[key] if key in definition else ''
                            if definition not in found and _value_matches_pattern(value, pattern, is_case, is_re):
                                found.add(definition)
                                yield definition
                if recursive:
                    object_collection += obj.definitions
            else:
                object_collection += obj.definitions
        elif isinstance(obj, Netlist):
            object_collection += obj.libraries
        elif isinstance(obj, Definition):
            if selection == Selection.INSIDE:
                for child in obj.children:
                    reference = child.reference
                    if reference:
                        if reference not in other_definitions:
                            other_definitions.add(reference)
                            if recursive:
                                object_collection.append(reference)
            else:
                for definition in obj.references:
                    parent = definition.parent
                    if parent:
                        if parent not in other_definitions:
                            other_definitions.add(parent)
                            if recursive:
                                object_collection.append(parent)
        elif isinstance(obj, Instance):
            if selection == Selection.INSIDE:
                reference = obj.reference
                if reference:
                    if reference not in other_definitions:
                        other_definitions.add(reference)
                        if recursive:
                            object_collection += reference.children
            else:
                parent = obj.parent
                if parent:
                    if parent not in other_definitions:
                        other_definitions.add(parent)
                        if recursive:
                            object_collection.append(parent)
        elif isinstance(obj, (Port, Cable)):
            definition = obj.definition
            if definition:
                if definition not in other_definitions:
                    other_definitions.add(definition)
        elif isinstance(obj, InnerPin):
            definition = obj.port
            if definition:
                object_collection.append(definition)
        elif isinstance(obj, OuterPin):
            definition = obj.instance
            if definition:
                object_collection.append(definition)
        elif isinstance(obj, Wire):
            cable = obj.cable
            if cable:
                object_collection.append(cable)
        elif isinstance(obj, HRef):
            if obj.is_valid:
                object_collection.append(obj.item)

    if other_definitions:
        namemap = dict()
        for other_definition in other_definitions:
            if other_definition in found:
                continue
            found.add(other_definition)
            name = other_definition[key] if key in other_definition else ''
            if name not in namemap:
                namemap[name] = list()
            namemap[name].append(other_definition)
        for pattern in patterns:
            pattern_is_absolute = _is_pattern_absolute(pattern, is_case, is_re)
            if pattern_is_absolute:
                if pattern in namemap:
                    result = namemap[pattern]
                    for definition in result:
                        yield definition
            else:
                names_to_remove = list()
                for name in namemap:
                    if _value_matches_pattern(name, pattern, is_case, is_re):
                        result = namemap[name]
                        names_to_remove.append(name)
                        for definition in result:
                            yield definition
                for name in names_to_remove:
                    del namemap[name]



import unittest
import spydrnet as sdn


class TestGetLibraries(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.netlist = sdn.Netlist()

        leaf_library = cls.netlist.create_library()
        leaf_library.name = 'primitives'

        library = cls.netlist.create_library()
        library.name = 'work'

        leaf_def = leaf_library.create_definition()
        leaf_def.name = 'leaf'
        leaf_port = leaf_def.create_port()
        leaf_port.name = 'I'
        leaf_port.create_pins(1)

        bottom_def = library.create_definition()
        bottom_def.name = 'bottom'
        bottom_port = bottom_def.create_port()
        bottom_port.name = 'I'
        bottom_port.create_pins(1)
        leaf_inst = bottom_def.create_child()
        leaf_inst.reference = leaf_def
        bottom_cable = bottom_def.create_cable()
        bottom_cable.name = 'bottom_cable'
        bottom_wire = bottom_cable.create_wire()
        bottom_wire.connect_pin(bottom_port.pins[0])
        bottom_wire.connect_pin(leaf_inst.pins[leaf_port.pins[0]])
        cls.leaf_inst = leaf_inst

        bottom_floating_wire = bottom_cable.create_wire()

        middle_def = library.create_definition()
        middle_def.name = 'middle'
        middle_port = middle_def.create_port()
        middle_port.name = "I"
        middle_port.create_pin()
        bottom_inst = middle_def.create_child()
        bottom_inst.name = 'bottom'
        bottom_inst.reference = bottom_def
        middle_cable = middle_def.create_cable()
        middle_cable.name = "middle_cable"
        middle_wire = middle_cable.create_wire()
        middle_wire.connect_pin(middle_port.pins[0])
        middle_wire.connect_pin(bottom_inst.pins[bottom_port.pins[0]])

        middle_floating_wire = middle_cable.create_wire()

        top_def = library.create_definition()
        top_def.name = 'top'
        top_port = top_def.create_port()
        top_port.name = "I"
        top_port.create_pin()
        middle_inst = top_def.create_child()
        middle_inst.name = 'middle'
        middle_inst.reference = middle_def
        top_cable = top_def.create_cable()
        top_cable.name = "top_cable"
        top_wire = top_cable.create_wire()
        top_wire.connect_pin(top_port.pins[0])
        top_wire.connect_pin(middle_inst.pins[middle_port.pins[0]])

        top_floating_wire = top_cable.create_wire()

        top_instance = sdn.Instance()
        top_instance.name = 'top'
        top_instance.reference = top_def
        cls.netlist.top_instance = top_instance

    def test_parameter_checking(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        library.name = "MY_LIB"
        self.assertRaises(TypeError, sdn.get_libraries, netlist, "MY_LIB", patterns="MY_LIB")
        self.assertRaises(TypeError, sdn.get_libraries, netlist, "MY_LIB", unsupported_keyword=None)
        self.assertRaises(TypeError, sdn.get_libraries, netlist, "MY_LIB", selection="BOTH")
        self.assertRaises(TypeError, sdn.get_libraries, None, "MY_LIB")
        self.assertRaises(TypeError, sdn.get_libraries, [None, netlist], "MY_LIB")

    def test_collection(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        library.name = "MY_LIB"
        ports = list(sdn.get_libraries([netlist, netlist]))
        self.assertEqual(len(ports), 1)

    def test_get_library_of_instances_outside(self):
        library_query = list(sdn.get_libraries(self.netlist.libraries[0].definitions[0].references,
                                               selection="OUTSIDE"))
        self.assertTrue(len(library_query) == 1 and library_query[0] == self.netlist.libraries[1])

    def test_get_library_of_instances_inside(self):
        library_query = list(sdn.get_libraries(self.netlist.libraries[0].definitions[0].references,
                                               selection="INSIDE"))
        self.assertTrue(len(library_query) == 1 and library_query[0] == self.netlist.libraries[0])

    def test_get_library_of_instances_recursive_down(self):
        library_query = list(sdn.get_libraries(self.netlist.top_instance, recursive=True))
        self.assertTrue(len(library_query) == 2 and all(x in library_query for x in self.netlist.libraries))

    def test_get_library_of_instances_recursive_up(self):
        library_query = list(sdn.get_libraries(self.leaf_inst, selection="OUTSIDE", recursive=True))
        self.assertTrue(len(library_query) == 1 and library_query[0] is self.netlist.libraries[1])

    def test_get_library_of_library_recursive_down(self):
        library_query = list(sdn.get_libraries(self.netlist.libraries[1], selection="INSIDE", recursive=True))
        self.assertTrue(len(library_query) == 2 and all(x in library_query for x in self.netlist.libraries))

    def test_get_library_of_library_recursive_up(self):
        library_query = list(sdn.get_libraries(self.netlist.libraries[0], selection="OUTSIDE", recursive=True))
        self.assertTrue(len(library_query) == 1 and library_query[0] is self.netlist.libraries[1])

    def test_get_library_of_inside_inside(self):
        library_query = list(sdn.get_libraries(self.netlist.libraries[0].definitions[0],
                                               selection="INSIDE"))
        self.assertTrue(len(library_query) == 1 and library_query[0] == self.netlist.libraries[0])

    def test_get_library_of_inside_outside(self):
        library_query = list(sdn.get_libraries(self.netlist.libraries[0].definitions[0],
                                               selection="OUTSIDE"))
        self.assertTrue(len(library_query) == 1 and library_query[0] == self.netlist.libraries[1])

    def test_get_library_of_inside_outside_recursive(self):
        library_query = list(sdn.get_libraries(self.netlist.libraries[0].definitions[0],
                                               selection="OUTSIDE", recursive=True))
        self.assertTrue(len(library_query) == 1 and library_query[0] == self.netlist.libraries[1])

    def test_get_library_of_inside_inside_recursive(self):
        library_query = list(sdn.get_libraries(self.netlist.top_instance.reference,
                                               selection="INSIDE", recursive=True))
        self.assertTrue(len(library_query) == 2 and all(x in library_query for x in self.netlist.libraries))

    def test_get_library_from_outer_pins(self):
        library_query = list(sdn.get_libraries(self.leaf_inst.pins))
        self.assertTrue(len(library_query) == 1 and library_query[0] == self.netlist.libraries[0])

    def test_get_library_from_inner_pins(self):
        library_query = list(sdn.get_libraries(self.leaf_inst.reference.ports[0].pins[0]))
        self.assertTrue(len(library_query) == 1 and library_query[0] == self.netlist.libraries[0])

    def test_get_library_from_wire(self):
        library_query = list(sdn.get_libraries(self.netlist.top_instance.reference.cables[0].wires[0]))
        self.assertTrue(len(library_query) == 1 and library_query[0] == self.netlist.libraries[1])

    def test_get_library_from_href(self):
        href = next(sdn.get_hinstances(self.netlist.top_instance))
        library_query = list(sdn.get_libraries(href))
        self.assertTrue(len(library_query) == 1 and library_query[0] == self.netlist.libraries[1])

    def test_unique_query_return(self):
        library_query = list(sdn.get_libraries([self.netlist, self.netlist.top_instance], "work"))
        self.assertTrue(len(library_query) == 1 and library_query[0] == self.netlist.libraries[1])

    def test_absolute_pattern_from_relative_query(self):
        library_query = list(sdn.get_libraries(self.netlist.top_instance, "work"))
        self.assertTrue(len(library_query) == 1 and library_query[0] == self.netlist.libraries[1])



from spydrnet import ir
from copy import deepcopy, copy, error

"""provide the clone function for spydrnet."""


def clone(element):
    """
    Clone any netlist objects

    several premises hold while cloning
     * the object will be orphaned and not belong to any parent
     * the object will maintain internal structure with cloned objects
     * the names will be unchanged
     * external connections will mostly be severed

    Properties
     * cloned using python's built in deepcopy functionality.
     * expected to be string objects but if you store something else there make sure you override deepcopy on that object.

    Instances have some special considerations
     * when cloned without the library containing the reference definition the instance will still point to the definition of it's clone.
     * in the same case as the above point the references of the definition will be updated accordingly
     * when a library is cloned some of the instances may be defined in another library these instances will follow the premises above
     * instances defined and referenced in the cloned library will point to the cloned definition

    """
    return element.clone()



import unittest

import spydrnet as sdn


class TestPrint(unittest.TestCase):

    netlist = sdn.Netlist(name='netlist')
    cable = sdn.Cable(name='cable', is_downto=False)
    instance = sdn.Instance()
    print(netlist)
    netlist.top_instance = instance
    print(netlist)
    print(cable)
    print(instance)
    library = netlist.create_library(name='lib')
    print(library)
    definition = sdn.Definition()
    print(definition)
    pin = sdn.Pin()
    cable = sdn.Cable()
    print(cable)
    wire = cable.create_wire()
    wire.connect_pin(pin)
    print(pin)
    print(wire)
    port = sdn.Port()
    port.direction = sdn.IN
    print(port)



from spydrnet import InnerPin, OuterPin, Wire, Netlist, Library, Definition, Instance, Port, Cable,\
    Element
from spydrnet.util.hierarchical_reference import HRef
from spydrnet.util.selection import Selection


def get_wires(obj, *args, **kwargs):
    """
    get_wires(obj, ...)

    Get wires *within* an object.

    Parameters
    ----------
    obj : object, Iterable - required
        The object or objects associated with this query. Queries return a collection of objects associated with the
        provided object or objects that match the query criteria. For example, `sdn.get_instances(netlist, ...)` would
        return all of the instances *within* the provided definition that match the additional criteria.
    recursive : bool - optional, default: False
        Specify if search should be recursive or not meaning that sub hierarchical instances within an instance are
        included or not.
    selection : Selection.{INSIDE, OUTSIDE, BOTH, ALL}, default: INSIDE
        This parameter determines the wires that are returned based on the instance associated with the object that is
        being searched.
    filter : function
        This is a single input function that can be used to filter out unwanted virtual instances. If not specifed, all
        matching virtual instances are returned. Otherwise, virtual instances that cause the filter function to evaluate
        to true are the only items returned.
    
    Returns
    -------
    wires : generator
        The wires associated with a particular object or collection of objects.
    
    """
    # Check argument list
    if len(args) > 0 or any(x not in {'selection', 'recursive', 'filter'} for x in
                            kwargs):
        raise TypeError("Unknown usage. Please see help for more information.")

    # Default values
    selection = kwargs.get('selection', Selection.INSIDE)
    if isinstance(selection, str):
        if selection in Selection.__members__:
            selection = Selection[selection]
    if isinstance(selection, Selection) is False:
        raise TypeError("selection must be '{}'".format("', '".join(Selection.__members__.keys())))

    filter_func = kwargs.get('filter', lambda x: True)
    recursive = kwargs.get('recursive', False)

    if isinstance(obj, (Element, HRef)) is False:
        try:
            object_collection = list(iter(obj))
        except TypeError:
            object_collection = [obj]
    else:
        object_collection = [obj]
    if all(isinstance(x, (HRef, Element)) for x in object_collection) is False:
        raise TypeError("get_hwires() supports all netlist related objects and hierarchical references or a "
                        "collection of theses as the object searched, unsupported object provided")

    return _get_wires(object_collection, selection, recursive, filter_func)


def _get_wires(object_collection, selection, recursive, filter_func):
    for result in filter(filter_func, _get_wires_raw(object_collection, selection, recursive)):
        yield result


def _get_wires_raw(object_collection, selection, recursive):
    in_yield = set()
    pin_search = set()
    while object_collection:
        obj = object_collection.pop()
        if isinstance(obj, Netlist):
            object_collection += obj.libraries
        elif isinstance(obj, Library):
            object_collection += obj.definitions
        elif isinstance(obj, Definition):
            if selection in {Selection.INSIDE, Selection.ALL}:
                # Get all cables inside a hierarchical instance
                for cable in obj.cables:
                    for wire in cable.wires:
                        if wire not in in_yield:
                            in_yield.add(wire)
                            yield wire
                # get internal cables recursively
                if recursive or selection == Selection.ALL:
                    object_collection += obj.children
            if selection in {Selection.OUTSIDE, Selection.BOTH, Selection.ALL}:
                for port in obj.ports:
                    for pin in port.pins:
                        pin_search.add(pin)
        elif isinstance(obj, Instance):
            reference = obj.reference
            if reference:
                object_collection.append(reference)
        elif isinstance(obj, Port):
            for pin in obj.pins:
                pin_search.add(pin)
        elif isinstance(obj, Cable):
            object_collection += obj.wires
        elif isinstance(obj, Wire):
            if selection == Selection.INSIDE:
                if obj not in in_yield:
                    in_yield.add(obj)
                    yield obj
            elif selection == Selection.OUTSIDE:
                for pin in obj.pins:
                    if isinstance(pin, OuterPin):
                        inner_wire = pin.inner_pin.wire
                        if inner_wire:
                            if inner_wire not in in_yield:
                                in_yield.add(inner_wire)
                                yield inner_wire
                    else:
                        cable = obj.cable
                        if cable:
                            definition = cable.definition
                            if definition:
                                for instance in definition.references:
                                    if pin in instance.pins:
                                        outer_pin = instance.pins[pin]
                                        outer_wire = outer_pin.wire
                                        if outer_wire:
                                            if outer_wire not in in_yield:
                                                in_yield.add(outer_wire)
                                                yield outer_wire
            else:
                for pin in obj.pins:
                    pin_search.add(pin)
        elif isinstance(obj, (InnerPin, OuterPin)):
            pin_search.add(obj)
        elif isinstance(obj, HRef):
            if obj.is_valid:
                object_collection.append(obj.item)

    while pin_search:
        new_pin_search = set()
        for pin in pin_search:
            if selection in {Selection.BOTH, Selection.ALL}:
                wire = pin.wire
                if wire and wire not in in_yield:
                    in_yield.add(wire)
                    yield wire
                    if selection == Selection.ALL:
                        for other_pin in wire.pins:
                            new_pin_search.add(other_pin)
                if isinstance(pin, OuterPin):
                    inner_pin = pin.inner_pin
                    if inner_pin:
                        inner_wire = inner_pin.wire
                        if inner_wire and inner_wire not in in_yield:
                            in_yield.add(inner_wire)
                            yield inner_wire
                            if selection == Selection.ALL:
                                for other_pin in inner_wire.pins:
                                    new_pin_search.add(other_pin)
                else:
                    port = pin.port
                    if port:
                        definition = port.definition
                        if definition:
                            for instance in definition.references:
                                outer_pin = instance.pins[pin]
                                outer_wire = outer_pin.wire
                                if outer_wire and outer_wire not in in_yield:
                                    in_yield.add(outer_wire)
                                    yield outer_wire
                                    if selection == Selection.ALL:
                                        for other_pin in outer_wire.pins:
                                            new_pin_search.add(other_pin)
            elif selection == Selection.INSIDE:
                if isinstance(pin, OuterPin):
                    inner_pin = pin.inner_pin
                    if inner_pin:
                        wire = inner_pin.wire
                        if wire and wire not in in_yield:
                            in_yield.add(wire)
                            yield wire
                else:
                    wire = pin.wire
                    if wire and wire not in in_yield:
                        in_yield.add(wire)
                        yield wire
            else:
                if isinstance(pin, OuterPin):
                    wire = pin.wire
                    if wire and wire not in in_yield:
                        in_yield.add(wire)
                        yield wire
                else:
                    port = pin.port
                    if port:
                        definition = port.definition
                        if definition:
                            for instance in definition.references:
                                outer_pin = instance.pins[pin]
                                outer_wire = outer_pin.wire
                                if outer_wire and outer_wire not in in_yield:
                                    in_yield.add(outer_wire)
                                    yield outer_wire
        pin_search = new_pin_search



_container_create_netlist = list()
_container_create_library = list()
_container_create_definition = list()
_container_create_port = list()
_container_create_cable = list()
_container_create_instance = list()
_container_cable_add_wire = list()
_container_cable_remove_wire = list()
_container_definition_add_port = list()
_container_definition_remove_port = list()
_container_definition_add_child = list()
_container_definition_remove_child = list()
_container_definition_add_cable = list()
_container_definition_remove_cable = list()
_container_instance_reference = list()
_container_library_add_definition = list()
_container_library_remove_definition = list()
_container_netlist_top_instance = list()
_container_netlist_add_library = list()
_container_netlist_remove_library = list()
_container_port_add_pin = list()
_container_port_remove_pin = list()
_container_wire_connect_pin = list()
_container_wire_disconnect_pin = list()
_container_dictionary_set = list()
_container_dictionary_delete = list()
_container_dictionary_pop = list()


def _call_create_netlist(*args, **kwargs):
    for func in _container_create_netlist:
        func(*args, **kwargs)


def _call_create_library(*args, **kwargs):
    for func in _container_create_library:
        func(*args, **kwargs)


def _call_create_definition(*args, **kwargs):
    for func in _container_create_definition:
        func(*args, **kwargs)


def _call_create_port(*args, **kwargs):
    for func in _container_create_port:
        func(*args, **kwargs)


def _call_create_cable(*args, **kwargs):
    for func in _container_create_cable:
        func(*args, **kwargs)


def _call_create_instance(*args, **kwargs):
    for func in _container_create_instance:
        func(*args, **kwargs)


def _call_cable_add_wire(*args, **kwargs):
    for func in _container_cable_add_wire:
        func(*args, **kwargs)


def _call_cable_remove_wire(*args, **kwargs):
    for func in _container_cable_remove_wire:
        func(*args, **kwargs)


def _call_definition_add_port(*args, **kwargs):
    for func in _container_definition_add_port:
        func(*args, **kwargs)


def _call_definition_remove_port(*args, **kwargs):
    for func in _container_definition_remove_port:
        func(*args, **kwargs)


def _call_definition_add_child(*args, **kwargs):
    for func in _container_definition_add_child:
        func(*args, **kwargs)


def _call_definition_remove_child(*args, **kwargs):
    for func in _container_definition_remove_child:
        func(*args, **kwargs)


def _call_definition_add_cable(*args, **kwargs):
    for func in _container_definition_add_cable:
        func(*args, **kwargs)


def _call_definition_remove_cable(*args, **kwargs):
    for func in _container_definition_remove_cable:
        func(*args, **kwargs)


def _call_instance_reference(*args, **kwargs):
    for func in _container_instance_reference:
        func(*args, **kwargs)


def _call_library_add_definition(*args, **kwargs):
    for func in _container_library_add_definition:
        func(*args, **kwargs)


def _call_library_remove_definition(*args, **kwargs):
    for func in _container_library_remove_definition:
        func(*args, **kwargs)


def _call_netlist_top_instance(*args, **kwargs):
    for func in _container_netlist_top_instance:
        func(*args, **kwargs)


def _call_netlist_add_library(*args, **kwargs):
    for func in _container_netlist_add_library:
        func(*args, **kwargs)


def _call_netlist_remove_library(*args, **kwargs):
    for func in _container_netlist_remove_library:
        func(*args, **kwargs)


def _call_port_add_pin(*args, **kwargs):
    for func in _container_port_add_pin:
        func(*args, **kwargs)


def _call_port_remove_pin(*args, **kwargs):
    for func in _container_port_remove_pin:
        func(*args, **kwargs)


def _call_wire_connect_pin(*args, **kwargs):
    for func in _container_wire_connect_pin:
        func(*args, **kwargs)


def _call_wire_disconnect_pin(*args, **kwargs):
    for func in _container_wire_disconnect_pin:
        func(*args, **kwargs)


def _call_dictionary_set(*args, **kwargs):
    for func in _container_dictionary_set:
        func(*args, **kwargs)


def _call_dictionary_delete(*args, **kwargs):
    for func in _container_dictionary_delete:
        func(*args, **kwargs)


def _call_dictionary_pop(*args, **kwargs):
    for func in _container_dictionary_pop:
        func(*args, **kwargs)


def register_create_netlist(method):
    _register(_container_create_netlist, method)


def register_create_library(method):
    _register(_container_create_library, method)


def register_create_definition(method):
    _register(_container_create_definition, method)


def register_create_port(method):
    _register(_container_create_port, method)


def register_create_cable(method):
    _register(_container_create_cable, method)


def register_create_instance(method):
    _register(_container_create_instance, method)


def register_cable_add_wire(method):
    _register(_container_cable_add_wire, method)


def register_cable_remove_wire(method):
    _register(_container_cable_remove_wire, method)


def register_definition_add_port(method):
    _register(_container_definition_add_port, method)


def register_definition_remove_port(method):
    _register(_container_definition_remove_port, method)


def register_definition_add_child(method):
    _register(_container_definition_add_child, method)


def register_definition_remove_child(method):
    _register(_container_definition_remove_child, method)


def register_definition_add_cable(method):
    _register(_container_definition_add_cable, method)


def register_definition_remove_cable(method):
    _register(_container_definition_remove_cable, method)


def register_instance_reference(method):
    _register(_container_instance_reference, method)


def register_library_add_definition(method):
    _register(_container_library_add_definition, method)


def register_library_remove_definition(method):
    _register(_container_library_remove_definition, method)


def register_netlist_top_instance(method):
    _register(_container_netlist_top_instance, method)


def register_netlist_add_library(method):
    _register(_container_netlist_add_library, method)


def register_netlist_remove_library(method):
    _register(_container_netlist_remove_library, method)


def register_port_add_pin(method):
    _register(_container_port_add_pin, method)


def register_port_remove_pin(method):
    _register(_container_port_remove_pin, method)


def register_wire_connect_pin(method):
    _register(_container_wire_connect_pin, method)


def register_wire_disconnect_pin(method):
    _register(_container_wire_disconnect_pin, method)


def register_dictionary_set(*args, **kwargs):
    _register(_container_dictionary_set, *args, **kwargs)


def register_dictionary_delete(*args, **kwargs):
    _register(_container_dictionary_delete, *args, **kwargs)


def register_dictionary_pop(*args, **kwargs):
    _register(_container_dictionary_pop, *args, **kwargs)


def _register(container_to_register, method):
    # TODO: look into inlining this function perhaps, not not be necessary since is won't be called often.
    assert method not in container_to_register
    container_to_register.append(method)


def deregister_create_netlist(method):
    _deregister(_container_create_netlist, method)


def deregister_create_library(method):
    _deregister(_container_create_library, method)


def deregister_create_definition(method):
    _deregister(_container_create_definition, method)


def deregister_create_port(method):
    _deregister(_container_create_port, method)


def deregister_create_cable(method):
    _deregister(_container_create_cable, method)


def deregister_create_instance(method):
    _deregister(_container_create_instance, method)


def deregister_cable_add_wire(method):
    _deregister(_container_cable_add_wire, method)


def deregister_cable_remove_wire(method):
    _deregister(_container_cable_remove_wire, method)


def deregister_definition_add_port(method):
    _deregister(_container_definition_add_port, method)


def deregister_definition_remove_port(method):
    _deregister(_container_definition_remove_port, method)


def deregister_definition_add_child(method):
    _deregister(_container_definition_add_child, method)


def deregister_definition_remove_child(method):
    _deregister(_container_definition_remove_child, method)


def deregister_definition_add_cable(method):
    _deregister(_container_definition_add_cable, method)


def deregister_definition_remove_cable(method):
    _deregister(_container_definition_remove_cable, method)


def deregister_instance_reference(method):
    _deregister(_container_instance_reference, method)


def deregister_library_add_definition(method):
    _deregister(_container_library_add_definition, method)


def deregister_library_remove_definition(method):
    _deregister(_container_library_remove_definition, method)


def deregister_netlist_top_instance(method):
    _deregister(_container_netlist_top_instance, method)


def deregister_netlist_add_library(method):
    _deregister(_container_netlist_add_library, method)


def deregister_netlist_remove_library(method):
    _deregister(_container_netlist_remove_library, method)


def deregister_port_add_pin(method):
    _deregister(_container_port_add_pin, method)


def deregister_port_remove_pin(method):
    _deregister(_container_port_remove_pin, method)


def deregister_wire_connect_pin(method):
    _deregister(_container_wire_connect_pin, method)


def deregister_wire_disconnect_pin(method):
    _deregister(_container_wire_disconnect_pin, method)


def deregister_dictionary_set(*args, **kwargs):
    _deregister(_container_dictionary_set, *args, **kwargs)


def deregister_dictionary_delete(*args, **kwargs):
    _deregister(_container_dictionary_delete, *args, **kwargs)


def deregister_dictionary_pop(*args, **kwargs):
    _deregister(_container_dictionary_pop, *args, **kwargs)


def _deregister(container_to_deregister, method):
    # TODO: look into inlining this function perhaps, may not be necessary since it won't be called often.
    assert method in container_to_deregister
    container_to_deregister.remove(method)



class GetterShortcuts:
    __slots__ = tuple()

    def get_netlists(self, *args, **kwargs):
        """get_netlists(...)

        Shortcut to :py:func:`~spydrnet.get_netlists`.
        """
        from spydrnet.util import get_netlists
        return get_netlists(self, *args, **kwargs)

    def get_libraries(self, *args, **kwargs):
        """get_libraries(...)

        Shortcut to :py:func:`~spydrnet.get_libraries`.
        """
        from spydrnet.util import get_libraries
        return get_libraries(self, *args, **kwargs)

    def get_definitions(self, *args, **kwargs):
        """get_definitions(...)

        Shortcut to :py:func:`~spydrnet.get_definitions`.
        """
        from spydrnet.util import get_definitions
        return get_definitions(self, *args, **kwargs)

    def get_instances(self, *args, **kwargs):
        """get_instances(...)

        Shortcut to :py:func:`~spydrnet.get_instances`.
        """
        from spydrnet.util import get_instances
        return get_instances(self, *args, **kwargs)

    def get_ports(self, *args, **kwargs):
        """get_ports(...)

        Shortcut to :py:func:`~spydrnet.get_ports`.
        """
        from spydrnet.util import get_ports
        return get_ports(self, *args, **kwargs)

    def get_pins(self, *args, **kwargs):
        """get_pins(...)

        Shortcut to :py:func:`~spydrnet.get_pins`.
        """
        from spydrnet.util import get_pins
        return get_pins(self, *args, **kwargs)

    def get_cables(self, *args, **kwargs):
        """get_cables(...)

        Shortcut to :py:func:`~spydrnet.get_cables`.
        """
        from spydrnet.util import get_cables
        return get_cables(self, *args, **kwargs)

    def get_wires(self, *args, **kwargs):
        """get_wires(...)

        Shortcut to :py:func:`~spydrnet.get_wires`.
        """
        from spydrnet.util import get_wires
        return get_wires(self, *args, **kwargs)

    def get_hinstances(self, *args, **kwargs):
        """get_hinstances(...)

        Shortcut to :py:func:`~spydrnet.get_hinstances`.
        """
        from spydrnet.util import get_hinstances
        return get_hinstances(self, *args, **kwargs)

    def get_hports(self, *args, **kwargs):
        """get_hports(...)

        Shortcut to :py:func:`~spydrnet.get_hports`.
        """
        from spydrnet.util import get_hports
        return get_hports(self, *args, **kwargs)

    def get_hpins(self, *args, **kwargs):
        """get_hpins(...)

        Shortcut to :py:func:`~spydrnet.get_hpins`.
        """
        from spydrnet.util import get_hpins
        return get_hpins(self, *args, **kwargs)

    def get_hcables(self, *args, **kwargs):
        """get_hcables(...)

        Shortcut to :py:func:`~spydrnet.get_hcables`.
        """
        from spydrnet.util import get_hcables
        return get_hcables(self, *args, **kwargs)

    def get_hwires(self, *args, **kwargs):
        """get_hwires(...)

        Shortcut to :py:func:`~spydrnet.get_hwires`.
        """
        from spydrnet.util import get_hwires
        return get_hwires(self, *args, **kwargs)



from spydrnet import FirstClassElement, InnerPin, OuterPin, Wire, Netlist, Library, Definition, Instance, Port, Cable
from spydrnet.util.hierarchical_reference import HRef
from spydrnet.util.selection import Selection
from spydrnet.util.patterns import _is_pattern_absolute, _value_matches_pattern


def get_hwires(obj, *args, **kwargs):
    """
    get_hwires(obj, ...)

    Get hierarchical references to wires *within* an object.

    Parameters
    ----------
    obj : object, Iterable - required
        The object or objects associated with this query. Queries return a collection of objects associated with the
        provided object or objects that match the query criteria. For example, `sdn.get_instances(netlist, ...)` would
        return all of the instances *within* the provided definition that match the additional criteria.
    patterns : str, Iterable - optional, positional or named, default: wildcard
        The search patterns. Patterns can be a single string or an Iterable collection of strings. Patterns can be
        absolute or they can contain wildcards or regular expressions. If `patterns` is not provided, then it defaults
        to a wildcard.
    is_case : bool - optional, named, default: True
        Specify if patterns should be treated as case sensitive. Only applies to patterns. Does not alter fast lookup
        behavior (if namespace policy uses case insensitive indexing, this parameter will not prevent a fast lookup
        from returning a matching object even if the case is not an exact match).
    is_re: bool - optional, named, default: False
        Specify if patterns are regular expressions. If `False`, a pattern can still contain `*` and `?` wildcards. A
        `*` matches zero or more characters. A `?` matches upto a single character.
    recursive : bool - optional, default: False
        Specify if search should be recursive or not meaning that sub hierarchical instances within an instance are
        included or not.
    selection : Selection.{INSIDE, OUTSIDE, BOTH, ALL}, default: INSIDE
        This parameter determines the wires that are returned based on the instance associated with the object that is
        being searched.
    filter : function
        This is a single input function that can be used to filter out unwanted virtual instances. If not specifed, all
        matching virtual instances are returned. Otherwise, virtual instances that cause the filter function to evaluate
        to true are the only items returned.
    
    Returns
    -------
    href_wires : generator
        The hierarchical references to wires associated with a particular object or collection of objects.
    
    """
    # Check argument list
    if len(args) == 1 and 'patterns' in kwargs:
        raise TypeError("get_hwires() got multiple values for argument 'patterns'")
    if len(args) > 1 or any(x not in {'patterns', 'selection', 'recursive', 'filter', 'is_case', 'is_re'} for x in
                            kwargs):
        raise TypeError("Unknown usage. Please see help for more information.")

    # Default values
    selection = kwargs.get('selection', Selection.INSIDE)
    if isinstance(selection, str):
        if selection in Selection.__members__:
            selection = Selection[selection]
    if isinstance(selection, Selection) is False:
        raise TypeError("selection must be '{}'".format("', '".join(Selection.__members__.keys())))

    filter_func = kwargs.get('filter', lambda x: True)
    recursive = kwargs.get('recursive', False)
    is_case = kwargs.get('is_case', True)
    is_re = kwargs.get('is_re', False)
    patterns = args[0] if len(args) == 1 else kwargs.get('patterns', ".*" if is_re else "*")

    if isinstance(obj, (FirstClassElement, InnerPin, OuterPin, Wire)) is False:
        try:
            object_collection = list(iter(obj))
        except TypeError:
            object_collection = [obj]
    else:
        object_collection = [obj]
    if all(isinstance(x, (HRef, FirstClassElement, InnerPin, OuterPin, Wire)) for x in object_collection) is False:
        raise TypeError("get_hwires() supports all netlist related objects and hierarchical references or a "
                        "collection of theses as the object searched, unsupported object provided")

    if isinstance(patterns, str):
        patterns = (patterns,)
    assert isinstance(patterns, (FirstClassElement, InnerPin, OuterPin, Wire)) is False

    return _get_hwires(object_collection, selection, patterns, recursive, is_case, is_re, filter_func)


def _get_hwires(object_collection, selection, patterns, recursive, is_case, is_re, filter_func):
    for result in filter(filter_func, _get_hwires_raw(object_collection, selection, patterns, recursive, is_case, is_re)):
        yield result


def _get_hwires_raw(object_collection, selection, patterns, recursive, is_case, is_re):
    in_namemap = set()
    in_yield = set()
    namemap = dict()
    hpin_search = set()
    bypass_namesearch = set()
    while object_collection:
        obj = object_collection.pop()
        if isinstance(obj, Netlist):
            top_instance = obj.top_instance
            if top_instance:
                href = HRef.from_parent_and_item(None, top_instance)
                object_collection.append(href)
        elif isinstance(obj, HRef):
            if obj.is_valid is False:
                continue
            item = obj.item
            if isinstance(item, Instance):
                if selection == Selection.INSIDE and obj not in bypass_namesearch:
                    _update_hwire_namemap(obj, recursive, in_namemap, namemap)
                else:
                    bypass_namesearch.discard(obj)
                    reference = item.reference
                    if reference:
                        if selection in {Selection.INSIDE, Selection.ALL}:
                            # Get all cables inside a hierarchical instance
                            for cable in reference.cables:
                                hcable = HRef.from_parent_and_item(obj, cable)
                                for wire in cable.wires:
                                    hwire = HRef.from_parent_and_item(hcable, wire)
                                    if hwire not in in_yield:
                                        in_yield.add(hwire)
                                        yield(hwire)
                            # get internal cables recursively
                            if recursive or selection == Selection.ALL:
                                for child in reference.children:
                                    href_child = HRef.from_parent_and_item(obj, child)
                                    bypass_namesearch.add(href_child)
                                    object_collection.append(href_child)
                        if selection in {Selection.OUTSIDE, Selection.BOTH, Selection.ALL}:
                            for port in reference.ports:
                                href_port = HRef.from_parent_and_item(obj, port)
                                for pin in port.pins:
                                    href_pin = HRef.from_parent_and_item(href_port, pin)
                                    hpin_search.add(href_pin)
            elif isinstance(item, Port):
                for pin in item.pins:
                    href_pin = HRef.from_parent_and_item(obj, pin)
                    hpin_search.add(href_pin)
            elif isinstance(item, Cable):
                for wire in item.wires:
                    href_wire = HRef.from_parent_and_item(obj, wire)
                    object_collection.append(href_wire)
            elif isinstance(item, Wire):
                if selection == Selection.INSIDE:
                    if obj not in in_yield:
                        in_yield.add(obj)
                        yield obj
                elif selection == Selection.OUTSIDE:
                    href_parent_cable = obj.parent
                    href_parent_instance = href_parent_cable.parent
                    for pin in item.pins:
                        if isinstance(pin, OuterPin):
                            href_inst = HRef.from_parent_and_item(href_parent_instance, pin.instance)
                            inner_wire = pin.inner_pin.wire
                            if inner_wire:
                                inner_cable = inner_wire.cable
                                href_cable = HRef.from_parent_and_item(href_inst, inner_cable)
                                href_wire = HRef.from_parent_and_item(href_cable, inner_wire)
                                if href_wire not in in_yield:
                                    in_yield.add(href_wire)
                                    yield href_wire
                        else:
                            href_parent = href_parent_instance.parent
                            if href_parent:
                                instance = href_parent_instance.item
                                if pin in instance.pins:
                                    outer_pin = instance.pins[pin]
                                    outer_wire = outer_pin.wire
                                    if outer_wire:
                                        outer_cable = outer_wire.cable
                                        href_cable = HRef.from_parent_and_item(href_parent, outer_cable)
                                        href_wire = HRef.from_parent_and_item(href_cable, outer_wire)
                                        if href_wire not in in_yield:
                                            in_yield.add(href_wire)
                                            yield href_wire
                else:
                    if obj not in in_yield:
                        in_yield.add(obj)
                        yield obj
                    href_cable = obj.parent
                    href_inst = href_cable.parent
                    for pin in item.pins:
                        if isinstance(pin, OuterPin):
                            href_sub_inst = HRef.from_parent_and_item(obj.parent, pin.instance)
                            inner_pin = pin.inner_pin
                            port = inner_pin.port
                            href_port = HRef.from_parent_and_item(href_sub_inst, port)
                            href_pin = HRef.from_parent_and_item(href_port, pin)
                        else:
                            port = pin.port
                            href_port = HRef.from_parent_and_item(href_inst, port)
                            href_pin = HRef.from_parent_and_item(href_port, pin)
                        object_collection.append(href_pin)
            elif isinstance(item, InnerPin):
                hpin_search.add(obj)
        elif isinstance(obj, Library):
            object_collection += obj.definitions
        elif isinstance(obj, Definition):
            hrefs = set(HRef.get_all_hrefs_of_instances(obj.references))
            bypass_namesearch |= hrefs
            object_collection += hrefs
        elif isinstance(obj, Instance):
            hrefs = set(HRef.get_all_hrefs_of_instances(obj))
            bypass_namesearch |= hrefs
            object_collection += hrefs
        elif isinstance(obj, (Port, Cable, InnerPin, OuterPin, Wire)):
            object_collection += HRef.get_all_hrefs_of_item(obj)

    if hpin_search:
        for hwire in _get_hwires_from_hpins(hpin_search, selection):
            if hwire not in in_yield:
                in_yield.add(hwire)
                yield hwire

    for href in in_yield:
        in_namemap.discard(href)

    if in_namemap:
        for pattern in patterns:
            pattern_is_absolute = _is_pattern_absolute(pattern, is_case, is_re)
            if pattern_is_absolute:
                if pattern in namemap:
                    result = namemap[pattern]
                    for href in result:
                        if href in in_namemap:
                            in_namemap.remove(href)
                            yield href
            else:
                for name in namemap:
                    if _value_matches_pattern(name, pattern, is_case, is_re):
                        result = namemap[name]
                        for href in result:
                            if href in in_namemap:
                                in_namemap.remove(href)
                                yield href


def _update_hwire_namemap(href_instance, recursive, found, namemap):
    search_stack = [(href_instance, False)]
    name_stack = list()
    while search_stack:
        href_instance, visited = search_stack.pop()
        if visited:
            name_stack.pop()
        else:
            search_stack.append((href_instance, True))
            name_stack.append(href_instance.item.name if href_instance.item.name else '')
            item = href_instance.item
            reference = item.reference
            if reference:
                for cable in reference.cables:
                    hcable = HRef.from_parent_and_item(href_instance, cable)
                    name_stack.append(cable.name if cable.name else '')
                    cable_hname = '/'.join(name_stack[1:])
                    for wire_index, wire in enumerate(cable.wires):
                        hwire = HRef.from_parent_and_item(hcable, wire)
                        if hwire not in found:
                            found.add(hwire)
                            if cable.is_scalar:
                                hname = cable_hname
                            else:
                                hname = "{}[{}]".format(cable_hname, cable.lower_index + wire_index)
                            if hname not in namemap:
                                namemap[hname] = list()
                            namemap[hname].append(hwire)
                    name_stack.pop()
                if recursive:
                    for child in reference.children:
                        if child.reference and child.reference.is_leaf() is False:
                            href_child = HRef.from_parent_and_item(href_instance, child)
                            search_stack.append((href_child, False))


def _get_hwires_from_hpins(hpin_search, selection):
    found_hwires = set()
    search_stack = list(hpin_search)
    while search_stack:
        hpin = search_stack.pop()
        if selection in {Selection.INSIDE, Selection.BOTH, Selection.ALL}:
            hwire_inside = _get_inner_hwire_from_hpin(hpin)
            if hwire_inside and hwire_inside not in found_hwires:
                found_hwires.add(hwire_inside)
                yield hwire_inside
                if selection is Selection.ALL:
                    search_stack += (x for x in _get_hpins_from_hwire(hwire_inside) if x != hpin)

        if selection in {Selection.OUTSIDE, Selection.BOTH, Selection.ALL}:
            hwire_outside = _get_outer_hwire_from_hpin(hpin)
            if hwire_outside and hwire_outside not in found_hwires:
                found_hwires.add(hwire_outside)
                yield hwire_outside
                if selection is Selection.ALL:
                    search_stack += (x for x in _get_hpins_from_hwire(hwire_outside) if x != hpin)


def _get_inner_hwire_from_hpin(hpin):
        wire = hpin.item.wire
        if wire:
            cable = wire.cable
            if cable:
                hport = hpin.parent
                hinst = hport.parent
                hcable = HRef.from_parent_and_item(hinst, cable)
                hwire = HRef.from_parent_and_item(hcable, wire)
                return hwire


def _get_outer_hwire_from_hpin(hpin):
    hport = hpin.parent
    hinst = hport.parent
    instance = hinst.item
    pin = hpin.item
    if pin in instance.pins:
        outer_pin = instance.pins[hpin.item]
        outer_wire = outer_pin.wire
        if outer_wire:
            cable = outer_wire.cable
            if cable:
                hcable = HRef.from_parent_and_item(hinst.parent, cable)
                hwire = HRef.from_parent_and_item(hcable, outer_wire)
                return hwire


def _get_hpins_from_hwire(hwire):
    hcable = hwire.parent
    hinst = hcable.parent
    for pin in hwire.item.pins:
        if isinstance(pin, InnerPin):
            port = pin.port
            if port:
                hport = HRef.from_parent_and_item(hinst, port)
                hpin = HRef.from_parent_and_item(hport, pin)
                yield hpin
        else:
            instance = pin.instance
            inner_pin = pin.inner_pin
            if instance and inner_pin:
                port = inner_pin.port
                if port:
                    other_hinst = HRef.from_parent_and_item(hinst, instance)
                    other_hport = HRef.from_parent_and_item(other_hinst, port)
                    other_hpin = HRef.from_parent_and_item(other_hport, inner_pin)
                    yield other_hpin



import unittest
import spydrnet as sdn
from spydrnet.composers.edif.edifify_names import EdififyNames


class TestEdififyNames(unittest.TestCase):

    def test_simple_conversion(self):
        ed = EdififyNames()
        l = list()
        i = sdn.Instance()
        i.name = "name"
        l = [i]
        assert ed.make_valid(i, l) == "name"

    def test_invalid_name(self):
        ed = EdififyNames()
        l = list()
        i = sdn.Instance()
        i.name = "*this_is+an$*`id[0:3]"
        l = [i]
        assert ed.make_valid(
            i, l) == "&_this_is_an___id_0_3_"

    def test_duplicate_name(self):
        ed = EdififyNames()
        l = list()
        i = sdn.Instance()
        i.name = 'name'
        i2 = sdn.Instance()
        i2.name = 'name'
        l = [i, i2]
        assert ed.make_valid(i2, l) == "name_sdn_1_"

    def test_duplicate_number_increment(self):
        ed = EdififyNames()
        l = list()
        i = sdn.Instance()
        i.name = 'name'
        i_duplicate = sdn.Instance()
        i_duplicate.name = 'name'
        i2 = sdn.Instance()
        i2.name = 'name_sdn_1_'
        l = [i, i2, i_duplicate]
        assert ed.make_valid(i_duplicate, l) == "name_sdn_2_"

    def test_length(self):
        ed = EdififyNames()
        l = list()
        test_str = ""
        answer_str = ""
        for _ in range(ed.name_length_target + 5):
            test_str = test_str + 'a'
        for _ in range(ed.name_length_target):
            answer_str = answer_str + 'a'
        i = sdn.Instance()
        i.name = test_str
        l = [i]
        assert ed.make_valid(i, l) == answer_str

    def test_length_2(self):
        ed = EdififyNames()
        l = list()
        test_str = "$"
        answer_str = "&_"
        for _ in range(ed.name_length_target + 5):
            test_str = test_str + 'a'
        for _ in range(ed.name_length_target - 2):
            answer_str = answer_str + 'a'
        i = sdn.Instance()
        i.name = test_str
        l = [i]
        assert ed.make_valid(i, l) == answer_str

    def test_length_3(self):
        ed = EdififyNames()
        l = list()

        instance_1 = sdn.Instance()
        name_1 = '&_'
        for _ in range(ed.name_length_target - 2):
            name_1 = name_1 + 'a'
        instance_1.name = name_1

        instance_2 = sdn.Instance()

        name_2 = '&_'
        for _ in range(ed.name_length_target - 9):
            name_2 = name_2 + 'a'
        name_2 = name_2 + '_sdn_1_'
        instance_2.name = name_2

        test_str = '$'
        answer_str = '&_'
        for _ in range(ed.name_length_target + 5):
            test_str = test_str + 'a'
        for _ in range(ed.name_length_target - 9):
            answer_str = answer_str + 'a'

        answer_str = answer_str + '_sdn_2_'

        instance_3 = sdn.Instance()
        instance_3.name = test_str

        l = [instance_1, instance_2, instance_3]

        assert ed.make_valid(instance_3, l) == answer_str



from spydrnet.global_state import global_callback


class CallbackListener:
    """
    extend this class to create a listener that can be used to make a plugin to spydrnet.
    In this class are all of the functions that are used as callbacks. There are also
    register methods whos function is to register the callback functions that are present.

    callbacks are netlist dependent. If a callback is registered to one netlist it is not automatically registered to all netlists.
    """

    def __init__(self):
        self.register_all_listeners()

    def create_netlist(self, netlist):
        raise NotImplementedError

    def create_library(self, library):
        raise NotImplementedError

    def create_definition(self, definition):
        raise NotImplementedError

    def create_port(self, port):
        raise NotImplementedError

    def create_cable(self, cable):
        raise NotImplementedError

    def create_instance(self, instance):
        raise NotImplementedError

    def cable_add_wire(self, cable, wire):
        raise NotImplementedError

    def cable_remove_wire(self, cable, wire):
        raise NotImplementedError

    def definition_add_port(self, definition, port):
        raise NotImplementedError

    def definition_remove_port(self, definition, port):
        raise NotImplementedError

    def definition_add_child(self, definition, child):
        raise NotImplementedError

    def definition_remove_child(self, definition, child):
        raise NotImplementedError

    def definition_add_cable(self, definition, cable):
        raise NotImplementedError

    def definition_remove_cable(self, definition, cable):
        raise NotImplementedError

    def instance_reference(self, instance, reference):
        raise NotImplementedError

    def library_add_definition(self, library, definition):
        raise NotImplementedError

    def library_remove_definition(self, library, definition):
        raise NotImplementedError

    def netlist_top_instance(self, netlist, instance):
        raise NotImplementedError

    def netlist_add_library(self, netlist, library):
        raise NotImplementedError

    def netlist_remove_library(self, netlist, library):
        raise NotImplementedError

    def port_add_pin(self, port, pin):
        raise NotImplementedError

    def port_remove_pin(self, port, pin):
        raise NotImplementedError

    def wire_connect_pin(self, wire, pin):
        raise NotImplementedError

    def wire_disconnect_pin(self, wire, pin):
        raise NotImplementedError
    
    def dictionary_set(self, element, key, value):
        raise NotImplementedError

    def dictionary_delete(self, element, key):
        raise NotImplementedError

    def dictionary_pop(self, element, key):
        raise NotImplementedError

    def register_all_listeners(self):
        if self.create_netlist.__func__ is not CallbackListener.create_netlist:
            self.register_create_netlist()

        if self.create_library.__func__ is not CallbackListener.create_library:
            self.register_create_library()

        if self.create_definition.__func__ is not CallbackListener.create_definition:
            self.register_create_definition()

        if self.create_port.__func__ is not CallbackListener.create_port:
            self.register_create_port()

        if self.create_cable.__func__ is not CallbackListener.create_cable:
            self.register_create_cable()

        if self.create_instance.__func__ is not CallbackListener.create_instance:
            self.register_create_instance()

        if self.cable_add_wire.__func__ is not CallbackListener.cable_add_wire:
            self.register_cable_add_wire()

        if self.cable_remove_wire.__func__ is not CallbackListener.cable_remove_wire:
            self.register_cable_remove_wire()

        if self.definition_add_port.__func__ is not CallbackListener.definition_add_port:
            self.register_definition_add_port()

        if self.definition_remove_port.__func__ is not CallbackListener.definition_remove_port:
            self.register_definition_remove_port()

        if self.definition_add_child.__func__ is not CallbackListener.definition_add_child:
            self.register_definition_add_child()

        if self.definition_remove_child.__func__ is not CallbackListener.definition_remove_child:
            self.register_definition_remove_child()

        if self.definition_add_cable.__func__ is not CallbackListener.definition_add_cable:
            self.register_definition_add_cable()

        if self.definition_remove_cable.__func__ is not CallbackListener.definition_remove_cable:
            self.register_definition_remove_cable()

        if self.instance_reference.__func__ is not CallbackListener.instance_reference:
            self.register_instance_reference()

        if self.library_add_definition.__func__ is not CallbackListener.library_add_definition:
            self.register_library_add_definition()

        if self.library_remove_definition.__func__ is not CallbackListener.library_remove_definition:
            self.register_library_remove_definition()

        if self.netlist_top_instance.__func__ is not CallbackListener.netlist_top_instance:
            self.register_netlist_top_instance()

        if self.netlist_add_library.__func__ is not CallbackListener.netlist_add_library:
            self.register_netlist_add_library()

        if self.netlist_remove_library.__func__ is not CallbackListener.netlist_remove_library:
            self.register_netlist_remove_library()

        if self.port_add_pin.__func__ is not CallbackListener.port_add_pin:
            self.register_port_add_pin()

        if self.port_remove_pin.__func__ is not CallbackListener.port_remove_pin:
            self.register_port_remove_pin()

        if self.wire_connect_pin.__func__ is not CallbackListener.wire_connect_pin:
            self.register_wire_connect_pin()

        if self.wire_disconnect_pin.__func__ is not CallbackListener.wire_disconnect_pin:
            self.register_wire_disconnect_pin()

        if self.dictionary_set.__func__ is not CallbackListener.dictionary_set:
            self.register_dictionary_set()

        if self.dictionary_delete.__func__ is not CallbackListener.dictionary_delete:
            self.register_dictionary_delete()

        if self.dictionary_pop.__func__ is not CallbackListener.dictionary_pop:
            self.register_dictionary_pop()
                
    def register_create_netlist(self):
        global_callback.register_create_netlist(self.create_netlist)

    def register_create_library(self):
        global_callback.register_create_library(self.create_library)

    def register_create_definition(self):
        global_callback.register_create_definition(self.create_definition)

    def register_create_port(self):
        global_callback.register_create_port(self.create_port)

    def register_create_cable(self):
        global_callback.register_create_cable(self.create_cable)

    def register_create_instance(self):
        global_callback.register_create_instance(self.create_instance)

    def register_cable_add_wire(self):
        global_callback.register_cable_add_wire(self.cable_add_wire)

    def register_cable_remove_wire(self):
        global_callback.register_cable_remove_wire(self.cable_remove_wire)

    def register_definition_add_port(self):
        global_callback.register_definition_add_port(self.definition_add_port)

    def register_definition_remove_port(self):
        global_callback.register_definition_remove_port(self.definition_remove_port)

    def register_definition_add_child(self):
        global_callback.register_definition_add_child(self.definition_add_child)

    def register_definition_remove_child(self):
        global_callback.register_definition_remove_child(self.definition_remove_child)

    def register_definition_add_cable(self):
        global_callback.register_definition_add_cable(self.definition_add_cable)

    def register_definition_remove_cable(self):
        global_callback.register_definition_remove_cable(self.definition_remove_cable)

    def register_instance_reference(self):
        global_callback.register_instance_reference(self.instance_reference)

    def register_library_add_definition(self):
        global_callback.register_library_add_definition(self.library_add_definition)

    def register_library_remove_definition(self):
        global_callback.register_library_remove_definition(self.library_remove_definition)

    def register_netlist_top_instance(self):
        global_callback.register_netlist_top_instance(self.netlist_top_instance)

    def register_netlist_add_library(self):
        global_callback.register_netlist_add_library(self.netlist_add_library)

    def register_netlist_remove_library(self):
        global_callback.register_netlist_remove_library(self.netlist_remove_library)

    def register_port_add_pin(self):
        global_callback.register_port_add_pin(self.port_add_pin)

    def register_port_remove_pin(self):
        global_callback.register_port_remove_pin(self.port_remove_pin)

    def register_wire_connect_pin(self):
        global_callback.register_wire_connect_pin(self.wire_connect_pin)

    def register_wire_disconnect_pin(self):
        global_callback.register_wire_disconnect_pin(self.wire_disconnect_pin)

    def register_dictionary_set(self):
        global_callback.register_dictionary_set(self.dictionary_set)

    def register_dictionary_delete(self):
        global_callback.register_dictionary_delete(self.dictionary_delete)

    def register_dictionary_pop(self):
        global_callback.register_dictionary_pop(self.dictionary_pop)

    def deregister_all_listeners(self):
        if self.create_netlist.__func__ is not CallbackListener.create_netlist:
            self.deregister_create_netlist()

        if self.create_library.__func__ is not CallbackListener.create_library:
            self.deregister_create_library()

        if self.create_definition.__func__ is not CallbackListener.create_definition:
            self.deregister_create_definition()

        if self.create_port.__func__ is not CallbackListener.create_port:
            self.deregister_create_port()

        if self.create_cable.__func__ is not CallbackListener.create_cable:
            self.deregister_create_cable()

        if self.create_instance.__func__ is not CallbackListener.create_instance:
            self.deregister_create_instance()

        if self.cable_add_wire.__func__ is not CallbackListener.cable_add_wire:
            self.deregister_cable_add_wire()

        if self.cable_remove_wire.__func__ is not CallbackListener.cable_remove_wire:
            self.deregister_cable_remove_wire()

        if self.definition_add_port.__func__ is not CallbackListener.definition_add_port:
            self.deregister_definition_add_port()

        if self.definition_remove_port.__func__ is not CallbackListener.definition_remove_port:
            self.deregister_definition_remove_port()

        if self.definition_add_child.__func__ is not CallbackListener.definition_add_child:
            self.deregister_definition_add_child()

        if self.definition_remove_child.__func__ is not CallbackListener.definition_remove_child:
            self.deregister_definition_remove_child()

        if self.definition_add_cable.__func__ is not CallbackListener.definition_add_cable:
            self.deregister_definition_add_cable()

        if self.definition_remove_cable.__func__ is not CallbackListener.definition_remove_cable:
            self.deregister_definition_remove_cable()

        if self.instance_reference.__func__ is not CallbackListener.instance_reference:
            self.deregister_instance_reference()

        if self.library_add_definition.__func__ is not CallbackListener.library_add_definition:
            self.deregister_library_add_definition()

        if self.library_remove_definition.__func__ is not CallbackListener.library_remove_definition:
            self.deregister_library_remove_definition()

        if self.netlist_top_instance.__func__ is not CallbackListener.netlist_top_instance:
            self.deregister_netlist_top_instance()

        if self.netlist_add_library.__func__ is not CallbackListener.netlist_add_library:
            self.deregister_netlist_add_library()

        if self.netlist_remove_library.__func__ is not CallbackListener.netlist_remove_library:
            self.deregister_netlist_remove_library()

        if self.port_add_pin.__func__ is not CallbackListener.port_add_pin:
            self.deregister_port_add_pin()

        if self.port_remove_pin.__func__ is not CallbackListener.port_remove_pin:
            self.deregister_port_remove_pin()

        if self.wire_connect_pin.__func__ is not CallbackListener.wire_connect_pin:
            self.deregister_wire_connect_pin()

        if self.wire_disconnect_pin.__func__ is not CallbackListener.wire_disconnect_pin:
            self.deregister_wire_disconnect_pin()

        if self.dictionary_set.__func__ is not CallbackListener.dictionary_set:
            self.deregister_dictionary_set()

        if self.dictionary_delete.__func__ is not CallbackListener.dictionary_delete:
            self.deregister_dictionary_delete()

        if self.dictionary_pop.__func__ is not CallbackListener.dictionary_pop:
            self.deregister_dictionary_pop()

    def deregister_create_netlist(self):
        global_callback.deregister_create_netlist(self.create_netlist)

    def deregister_create_library(self):
        global_callback.deregister_create_library(self.create_library)

    def deregister_create_definition(self):
        global_callback.deregister_create_definition(self.create_definition)

    def deregister_create_port(self):
        global_callback.deregister_create_port(self.create_port)

    def deregister_create_cable(self):
        global_callback.deregister_create_cable(self.create_cable)

    def deregister_create_instance(self):
        global_callback.deregister_create_instance(self.create_instance)

    def deregister_cable_add_wire(self):
        global_callback.deregister_cable_add_wire(self.cable_add_wire)

    def deregister_cable_remove_wire(self):
        global_callback.deregister_cable_remove_wire(self.cable_remove_wire)

    def deregister_definition_add_port(self):
        global_callback.deregister_definition_add_port(self.definition_add_port)

    def deregister_definition_remove_port(self):
        global_callback.deregister_definition_remove_port(self.definition_remove_port)

    def deregister_definition_add_child(self):
        global_callback.deregister_definition_add_child(self.definition_add_child)

    def deregister_definition_remove_child(self):
        global_callback.deregister_definition_remove_child(self.definition_remove_child)

    def deregister_definition_add_cable(self):
        global_callback.deregister_definition_add_cable(self.definition_add_cable)

    def deregister_definition_remove_cable(self):
        global_callback.deregister_definition_remove_cable(self.definition_remove_cable)

    def deregister_instance_reference(self):
        global_callback.deregister_instance_reference(self.instance_reference)

    def deregister_library_add_definition(self):
        global_callback.deregister_library_add_definition(self.library_add_definition)

    def deregister_library_remove_definition(self):
        global_callback.deregister_library_remove_definition(self.library_remove_definition)

    def deregister_netlist_top_instance(self):
        global_callback.deregister_netlist_top_instance(self.netlist_top_instance)

    def deregister_netlist_add_library(self):
        global_callback.deregister_netlist_add_library(self.netlist_add_library)

    def deregister_netlist_remove_library(self):
        global_callback.deregister_netlist_remove_library(self.netlist_remove_library)

    def deregister_port_add_pin(self):
        global_callback.deregister_port_add_pin(self.port_add_pin)

    def deregister_port_remove_pin(self):
        global_callback.deregister_port_remove_pin(self.port_remove_pin)

    def deregister_wire_connect_pin(self):
        global_callback.deregister_wire_connect_pin(self.wire_connect_pin)

    def deregister_wire_disconnect_pin(self):
        global_callback.deregister_wire_disconnect_pin(self.wire_disconnect_pin)

    def deregister_dictionary_set(self):
        global_callback.deregister_dictionary_set(self.dictionary_set)

    def deregister_dictionary_delete(self):
        global_callback.deregister_dictionary_delete(self.dictionary_delete)

    def deregister_dictionary_pop(self):
        global_callback.deregister_dictionary_pop(self.dictionary_pop)


from spydrnet.ir.element import Element


class Pin(Element):
    """Pin connects to a single wire.

    This class is extended by InnerPin and OuterPin"""
    __slots__ = ['_wire', '__weakref__']

    def __init__(self):
        self._wire = None

    @property
    def wire(self):
        """Get the wire the pin is connected to. This value cannot be modified directly by the end user."""
        return self._wire

    def __str__(self):
        """Re-define the print function so it is easier to read"""
        rep = str(type(self))
        rep = rep[:-1] + '; '
        if self.wire is None:
            rep += 'Wire connected undefined'
        else:
            rep += 'connected to\'' + str(self.wire) + '\''
        rep += '>'
        return rep


import os
import zipfile
import tempfile

"""
Init for Spydrnet. the funcitons below can be called directly

"""


def parse(filename):
    """
    The parse function is able to parse either a EDIF (.edf) file or a Verilog file (.v)

    This functions also supports the parsing of .zip files. Such as the ones in support_files folder

    Returns
    -------
    Netlist
        The netlist that comes as the result of the parsing of the file if the file was parsed successfully

    Examples
    --------

    >>> import spydrnet as sdn
    >>> netlist = sdn.parse('<netlist_filename>.edf')

    Or we can parse a verilog file

    >>> netlist = sdn.parse('<netlist_filename>.v')

    Or a zip file that contains the edif or verilog file

    >>> netlist = sdn.parse('4bitadder.edf.zip')
    """
    basename_less_final_extension = os.path.splitext(
        os.path.basename(filename))[0]
    extension = get_lowercase_extension(filename)
    if extension == ".zip":
        assert zipfile.is_zipfile(filename), \
            "Input filename {} with extension .zip is not a zip file.".format(
                basename_less_final_extension)
        with tempfile.TemporaryDirectory() as tempdirname:
            with zipfile.ZipFile(filename) as zip:
                files = zip.namelist()
                assert len(files) == 1 and files[0] == basename_less_final_extension, \
                    "Only single file archives allowed with a file whose name matches the name of the archive"
                zip.extract(basename_less_final_extension, tempdirname)
                filename = os.path.join(
                    tempdirname, basename_less_final_extension)
                return _parse(filename)
    return _parse(filename)


def _parse(filename):
    extension = get_lowercase_extension(filename)
    if extension in [".edf", ".edif", ".edn"]:
        from spydrnet.parsers.edif.parser import EdifParser
        parser = EdifParser.from_filename(filename)
    elif extension in [".v", ".vh"]:
        from spydrnet.parsers.verilog.parser import VerilogParser
        parser = VerilogParser.from_filename(filename)
    else:
        raise RuntimeError("Extension {} not recognized.".format(extension))
    parser.parse()
    return parser.netlist


def get_lowercase_extension(filename):
    extension = os.path.splitext(filename)[1]
    extension_lower = extension.lower()
    return extension_lower



from spydrnet.ir.first_class_element import FirstClassElement
from spydrnet.ir.outerpin import OuterPin
from spydrnet.ir.views.outerpinsview import OuterPinsView
from spydrnet.global_state import global_callback
from spydrnet.global_state.global_callback import _call_create_instance
from copy import deepcopy, copy, error
from collections import OrderedDict


class Instance(FirstClassElement):
    """Netlist instance of a netlist definition. 

    Instances are literally instances of definitions and they reside inside other definitions.
    Function names have been set to prevent potential confusion that could arise because instances have both a parent definition and definitions which they reference.


    :ivar parent: the parent of the object. Parent is the definition that instances another definition.
    :ivar child: the instance itself is the child of the parent.
    :ivar reference: the item of the object. Reference is the definition of the instance.

    For example, when writing definition 1, we instance definition 2. Definition 1 is the parent, the instance is the child, and definition 2 is the instance's reference.

    """
    __slots__ = ['_parent', '_reference', '_pins','_is_top_instance']

    def __init__(self, name=None, properties=None):
        """Creates an empty object of type instance.

        parameters
        ----------

        name - (str) the name of this instance
        properties - (dict) the dictionary which holds the properties
        """
        super().__init__()
        self._parent = None
        self._reference = None
        self._pins = OrderedDict()
        self._is_top_instance = False
        _call_create_instance(self)
        if name != None:
            self.name = name
        if properties != None:
            assert isinstance(
                properties, dict), "properties must be a dictionary"
            for key in properties:
                self[key] = properties[key]

    @property
    def parent(self):
        """Get the definition that contains this instance"""
        return self._parent

    def test(self):

        return True

    @property
    def reference(self):
        """Get the definition that this instance is instantiating"""
        return self._reference

    @reference.setter
    def reference(self, value):
        """Change the definition that represents this instance.
        Port positioning and size must be taken into account when a new definition is being used.
        if they are different the connections cannot be done automatically with this function.

        parameters
        ----------

        value - (Definition) the definition that this instance should be an instance of"""
        global_callback._call_instance_reference(self, value)
        if value is None:
            for pin in self.pins:
                wire = pin.wire
                if wire:
                    wire.disconnect_pin(pin)
                if isinstance(pin, OuterPin):
                    pin._instance = None
                    pin._inner_pin = None
            self._pins.clear()
            if self._reference:
                self._reference._references.remove(self)
        else:
            if self._reference is not None:
                assert len(self.reference.ports) == len(value.ports) and all(len(x.pins) == len(y.pins) for x, y in
                                                                             zip(self.reference.ports, value.ports)), \
                    "Reference reassignment only supported for definitions with matching port positions"
                self._reference._references.remove(self)
                for cur_port, new_port in zip(self.reference.ports, value.ports):
                    for cur_pin, new_pin in zip(cur_port.pins, new_port.pins):
                        outer_pin = self._pins.pop(cur_pin)
                        outer_pin._inner_pin = new_pin
                        self._pins[new_pin] = outer_pin
            else:
                for port in value.ports:
                    for pin in port.pins:
                        self._pins[pin] = OuterPin.from_instance_and_inner_pin(
                            self, pin)
            value._references.add(self)
        self._reference = value

    @reference.deleter
    def reference(self):
        self.reference = None

    def get_ports(self, *args, **kwargs):
        from spydrnet.util.get_ports import get_ports
        return get_ports(self, *args, **kwargs)

    @property
    def pins(self):
        """Get the pins on this instance.
        Can do instance.pins[<inner_pin>] to get the inner pin's associated outer pins."""
        return OuterPinsView(self._pins)

    def _clone_rip_and_replace_in_definition(self, memo):
        """Slide the outerpins references into a new context. 

        The instance still references something outside of what has been cloned.
        """
        for op in self._pins.values():
            op._clone_rip_and_replace(memo)

    def _clone_rip_and_replace_in_library(self, memo):
        """Move the instance into a new library/netlist. 

        This will replace the reference if affected and replace the inner pins that will be affected as well.
        The instance should not be in the references list of the reference definition
        """
        new_pins = OrderedDict()
        for ip, op in self._pins.items():
            new_pins[memo[ip]] = op
        self._pins = new_pins

    def _clone_rip(self):
        """Remove the instance from its current environmnet. 

        This will remove the instance from any wires but it will add it in to the references set on the definition which it instantiates.
        """
        for op in self._pins.values():
            op._wire = None
        self._reference._references.add(self)

    def _clone(self, memo):
        """Not api safe clone function
        clone the instance leaving all references in tact.
        The instance can then either be ripped or ripped and replaced
        """
        assert self not in memo, "the object should not have been copied twice in this pass"
        c = Instance()
        memo[self] = c
        c._parent = None
        for inner_pin, outer_pin in self._pins.items():
            new_outer_pin = outer_pin._clone(memo)
            new_outer_pin._instance = c
            c._pins[inner_pin] = new_outer_pin
        c._reference = self._reference
        c._data = deepcopy(self._data)
        return c

    def clone(self):
        """

        Clone the instance in an api safe way.
        This call will return a cloned instance that has the following properties:

         * the pins in the instance will all be disconnected from wires but they will maintain their references to inner pins
         * the instance references is the same as the cloned object
         * the reference's references list contains this instance
         * the instance is orphaned (no longer a child of the definition to which the cloned definition belonged

        """
        c = self._clone(OrderedDict())
        c._clone_rip()
        return c

    def is_leaf(self):
        """Check to see if the definition that this instance contains represents a leaf cell.

        Leaf cells are cells with no children instances or no children cables. 
        Blackbox cells are considered leaf cells as well as direct pass through cells with cables only
        """
        if self._reference is None:
            return False
        elif len(self._reference._children) > 0 or len(self._reference._cables) > 0:
            return False
        return True

    def is_unique(self):
        """
        Check to see if the instance is unique
        """
        if len(self.reference.references) == 1 or self.reference.is_leaf():
            return True
        else:
            return False

    def __str__(self):
        """Re-define the print function so it is easier to read"""
        rep = super().__str__()
        rep = rep[:-1] + '; '
        if self.parent is None:
            rep += 'parent definition undefined'
        elif self.parent.name is None:
            rep += 'parent definition.name undefined'
        else:
            rep += 'parent definition.name \'' + self.parent.name + '\''

        rep += '; '

        if self.reference is None:
            rep += 'reference definition undefined'
        elif self.reference.name is None:
            rep += 'reference definition.name undefined'
        else:
            rep += 'reference definition.name \'' + self.reference.name + '\''
        rep += '>'
        return rep

    @property
    def is_top_instance(self):
        return self._is_top_instance

    @is_top_instance.setter
    def is_top_instance(self,value):
        self._is_top_instance = value


import unittest
import spydrnet as sdn
from spydrnet.parsers.verilog.verilog_token_factory import TokenFactory

# these code snippets may have copyright protection other than what the License file entails please use with caution

very_simple = '''// Copyright 1986-2020 Xilinx, Inc. All Rights Reserved.
// --------------------------------------------------------------------------------
// Tool Version: Vivado v.2020.2 (lin64) Build 3064766 Wed Nov 18 09:12:47 MST 2020
// Date        : Wed Mar 17 14:51:06 2021
// Host        : DMachine running 64-bit Linux Mint 20.1
// Command     : write_verilog ediftoverilog
// Design      : synth_th1_slaac
// Purpose     : This is a Verilog netlist of the current design or from a specific cell of the design. The output is an
//               IEEE 1364-2001 compliant Verilog HDL file that contains netlist information obtained from the input
//               design files.
// Device      : xc7k70tfbv676-1
// --------------------------------------------------------------------------------
`timescale 1 ps / 1 ps

(* STRUCTURAL_NETLIST = "yes" *) 
module synth_th1_slaac
   (XP_IN);
  input [7:0]XP_IN;


endmodule'''

very_simple_multi_line = '''/* Copyright 1986-2020 Xilinx, Inc. All Rights Reserved.
// --------------------------------------------------------------------------------
// Tool Version: Vivado v.2020.2 (lin64) Build 3064766 Wed Nov 18 09:12:47 MST 2020
// Date        : Wed Mar 17 14:51:06 2021
// Host        : DMachine running 64-bit Linux Mint 20.1
// Command     : write_verilog ediftoverilog
// Design      : synth_th1_slaac
// Purpose     : This is a Verilog netlist of the current design or from a specific cell of the design. The output is an
//               IEEE 1364-2001 compliant Verilog HDL file that contains netlist information obtained from the input
//               design files.
// Device      : xc7k70tfbv676-1
// --------------------------------------------------------------------------------
*/
`timescale 1 ps / 1 ps

(* STRUCTURAL_NETLIST = "yes" *) 
module synth_th1_slaac
   (XP_IN);
  input [7:0]XP_IN;


endmodule'''

port_remap = '''
(* STRUCTURAL_NETLIST = "yes" *)
module synth_th1_slaac
   (.XP_IN({\^XP_IN [0],\^XP_IN [1],\^XP_IN [2],\^XP_IN [3],\^XP_IN [4],\^XP_IN [5],\^XP_IN [6],\^XP_IN [7]}));

  input [7:0]\^XP_IN ;

  wire [7:0]\^XP_IN ;
  
endmodule
'''

comments = '''
/*/ this should be
treated
as a single comment /*/

//* this should be a single line comment
these should not be part of the comment

/*

multi line with extra white space


*/
'''

final_comments = '''
module synth_th1_slaac
   (XP_IN);
  input [7:0]XP_IN;

endmodule
//* Final design comments
'''


class TestVerilogParser(unittest.TestCase):

    def run_token_count(self, string):
        tf = TokenFactory()
        running_total = 0

        for character in string:
            token = tf.add_character(character)
            if token is not None:
                running_total += 1
        token = tf.flush()
        if token is not None:
            running_total += 1
        return running_total

    def test_counts(self):
        to_run = [(very_simple_multi_line, 24, "very_simple_multi_line"), (very_simple, 35, "very_simple"),
                  (port_remap, 74, "port_remap"), (comments, 11, "comments"), (final_comments, 16, "port_remap")]
        for p in to_run:
            s, c, n = p
            assert c == self.run_token_count(
                s), "the number of expected tokens did not match up with the number parsed " + n


if __name__ == '__main__':
    unittest.main()



from spydrnet.ir.pin import Pin
from copy import deepcopy, copy, error


class InnerPin(Pin):
    """Pins that correspond to definitions.

    These pins can be thought of as on the inside of a definition. There can be
    many outer pins for each inner pin
    """
    __slots__ = ['_port']

    def __init__(self):
        super().__init__()
        self._port = None

    @property
    def port(self):
        """Return the port that the inner pin is a part of.

        This object cannot be modified directly by the end user."""
        return self._port

    def _clone_rip_and_replace(self, memo):
        """Remove from its current environment and place it into the new cloned environment with references held in the memo dictionary"""
        if self._wire != None:
            assert self._wire in memo, "wire must have been cloned in order to rip and replace innerpin"
            self._wire = memo[self._wire]

    def _clone_rip(self):
        """Remove from its current environmnet. This will remove all pin pointers and create a floating stand alone instance."""
        self._wire = None

    def _clone(self, memo):
        """Not api safe clone function.

        Clone leaving all references in tact.
        the element can then either be ripped or ripped and replaced"""
        assert self not in memo, "the object should not have been copied twice in this pass"
        c = InnerPin()
        memo[self] = c
        c._wire = self._wire
        c._port = None
        return c

    def clone(self):
        """Clone the inner pin in an api safe way.

        The following conditions will be met:

         * The inner pin will be orphaned from any ports
         * The pin will not be connected to any wires
         * The pin will not be referenced to by any wires or outer pins

        """
        c = self._clone(dict())
        c._clone_rip()
        return c
